/**
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * This file is part of terrayazilim-json project.
 *
 * This file incorporates work covered by
 * the following copyright and permission notices:
 *
 * Copyright (C) 2018 Terra Yazılım Bilişim Hiz. Elek. Dan. Oto. ve Loj. Tic. Ltd. Şti. | info [at] terrayazilim [dot] com [dot] tr
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package tr.com.terrayazilim.json;

import tr.com.terrayazilim.json.stream.JsonBuffer;
import java.lang.reflect.Field;
import java.lang.reflect.Type;
import java.math.BigDecimal;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import tr.com.terrayazilim.core.Objects;
import tr.com.terrayazilim.core.Strings;
import tr.com.terrayazilim.json.codec.JsonArrayAdapter;
import tr.com.terrayazilim.json.codec.JsonSerializerContext;
import tr.com.terrayazilim.json.codec.JsonCollectionAdapter;
import tr.com.terrayazilim.json.codec.JsonDeserializerContext;
import tr.com.terrayazilim.json.codec.JsonMapAdapter;
import tr.com.terrayazilim.json.codec.JsonNumberAdapter;
import tr.com.terrayazilim.json.type.Token;
import tr.com.terrayazilim.json.stream.JsonStream;
import tr.com.terrayazilim.json.type.ConstructorToken;
import tr.com.terrayazilim.json.type.FieldToken;

/**
 * @see tr.com.terrayazilim.json.JsonPolicy
 * @see tr.com.terrayazilim.json.JsonBuilder
 *
 * <p>
 * Json instances are thread safe. Instances can be used across multiple threads. Currently, we
 * support adapters that listed in below as default.
 *
 * <ul>
 * <li> Primitive Types ({@code boolean}, {@code int}, {@code long} etc.) </li>
 * <li> Wrapper Types ({@code Boolean}, {@code Integer}, {@code Long} etc.) </li>
 * <li> {@code java.util.Date} Type </li>
 * <li> {@code java.util.Calendar} Type </li>
 * <li> Both Primitive and Generic Arrays </li>
 * <li> Collection Types ({@code java.util.List}, {@code java.util.Set} etc) </li>
 * <li> Map Types ({@code java.util.HashMap}, {@code java.util.LinkedHashMap} etc) </li>
 * <li> Both Enum Serialization and Deserialization </li>
 * <li> Nested Objects </li>
 * </ul>
 * </p>
 *
 * <p>
 * See the {@link tr.com.terrayazilim.json.JsonPolicy} for default configuration options.
 * </p>
 *
 * <p>
 * Usage notes: Json can serialize and deserialize generic and wildcard implementations. However, we
 * highly suggest avoid using hard-coded({@literal <?>}) wildcard types.
 * </p>
 *
 * <p>
 * Error Codes definitions and possible solutions:
 *
 * {@literal 101, 102}, Possibly forced JsonAdapter. Register {@code JsonAdapter} from
 * {@code JsonBuilder} or {@code JsonPolicy}.
 *
 * {@literal 103, 104}, Possibly forced JsonSerializer or JsonDeserializer. See ErrorCode:101,
 * ErrorCode:102.
 *
 * {@literal 105}, Invalid JSON String. {@literal 106}, Not registered and/or not recognizable
 * Object Type. {@literal 107}, Deserialization error. It occurs when the bean doesn't have default
 * constructor. Either create a default constructor in that bean or use {@code JsonDeserializer}.
 * {@literal 108}, Annotated Adapter doesn't have default, non-arg constructor. {@literal 109},
 * Annotated Serializer doesn't have default, non-arg constructor. {@literal 110}, Annotated
 * Deserializer doesn't have default, non-arg constructor. {@literal 111}, Arg of JsonLiteral
 * Constructor is not acceptable. See {@link tr.com.terrayazilim.json.JsonLiteral #isLiteral}
 * {@literal 112}, Status char is not any of [i, a, o , d, k]. This should never happen.
 * {@literal 113}, Parameter of JsonStream Constructor#JsonValue is not instance any of [JsonNull,
 * JsonLiteral, JsonObject, JsonArray]. It may occur when the hard-coded JsonValue trying to
 * register in JsonStream. {@literal 114}, See the ErrorCode:113. {@literal 115}, Unexpected
 * JsonStream Status. This should never happen. See the ErrorCode:113. {@literal 116,117,118,119},
 * JsonStream next# casting format problem.
 *
 * </p>
 *
 * @author M.Çağrı Tepebaşılı - cagritepebasili [at] protonmail [dot] com
 * @version 1.0.3
 * @version 1.0.3 ~ 1.0.0:JSON.java
 */
public final class Json {

  private final ThreadLocal<Map<Token<?>, JsonRuntimeAdapter<?>>> adapters;
  private final Map<Token<?>, JsonAdapter<?>> cache;
  private final Map<Token<?>, JsonAdapter<?>> proxy;

  private final Map<String, JsonRuntimeSerializer<?>> serializers;
  private final Map<String, JsonRuntimeDeserializer<?>> deserializers;
  private final JsonPolicy policy;

  /**
   * @param builder
   */
  public Json(JsonBuilder builder) {
    this(builder.getPolicy());
  }

  public Json() {
    this(new JsonPolicy());
  }

  /**
   * @param policy
   */
  public Json(JsonPolicy policy) {
    Map<Token<?>, JsonAdapter<?>> stack = new HashMap<>();

    includeCharAdapters(stack);
    includeNumberAdapters(stack);
    includeBoolAdapters(stack);
    includeJsonAdapters(stack);
    includeTimeAdapters(stack);

    stack.put(new Token<>(UUID.class), getUUIDAdapter());
    stack.put(NULL_TOKEN, getNullAdapter());

    this.policy = policy;
    embedAdapters(policy.getAdapters(), stack);

    Map<String, JsonRuntimeSerializer<?>> serializerStack = new HashMap<>();
    embedSerializers(policy.getSerializers(), stack, serializerStack);

    Map<String, JsonRuntimeDeserializer<?>> deserializerStack = new HashMap<>();
    embedDeserializers(policy.getDeserializers(), stack, deserializerStack);

    this.serializers = Collections.unmodifiableMap(serializerStack);
    this.deserializers = Collections.unmodifiableMap(deserializerStack);

    this.proxy = Collections.unmodifiableMap(stack);
    this.cache = new ConcurrentHashMap<>();
    this.adapters = new ThreadLocal<>();
  }

  /**
   * Adds JsonType adapters.
   *
   * @param stack
   */
  private void includeJsonAdapters(Map<Token<?>, JsonAdapter<?>> stack) {
    stack.put(new Token<JsonLiteral>(JsonLiteral.class), getJsonLiteralAdapter());
    stack.put(new Token<JsonArray>(JsonArray.class), getJsonArrayAdapter());
    stack.put(new Token<JsonObject>(JsonObject.class), getJsonObjectAdapter());
    stack.put(new Token<JsonNull>(JsonNull.class), getJsonNullAdapter());
  }

  /**
   * Adds Number primitive adapters. BigDecimal included.
   *
   * @param stack
   */
  private void includeNumberAdapters(Map<Token<?>, JsonAdapter<?>> stack) {
    stack.put(new Token<BigDecimal>(BigDecimal.class), getBigDecimalAdapter());
    stack.put(new Token<Number>(Number.class), new JsonNumberAdapter());
    stack.put(new Token<Double>(double.class), getDoubleadapter());
    stack.put(new Token<Double>(Double.class), getDoubleadapter());
    stack.put(new Token<Integer>(int.class), getIntAdapter());
    stack.put(new Token<Integer>(Integer.class), getIntAdapter());
    stack.put(new Token<Short>(short.class), getShortAdapter());
    stack.put(new Token<Short>(Short.class), getShortAdapter());
    stack.put(new Token<Byte>(Byte.class), getByteAdapter());
    stack.put(new Token<Byte>(byte.class), getByteAdapter());
    stack.put(new Token<Float>(Float.class), getFloatAdapter());
    stack.put(new Token<Float>(float.class), getFloatAdapter());
    stack.put(new Token<Long>(long.class), getLongAdapter());
    stack.put(new Token<Long>(Long.class), getLongAdapter());
  }

  /**
   * Adds Character based class adapters.
   *
   * @param stack
   */
  private void includeCharAdapters(Map<Token<?>, JsonAdapter<?>> stack) {
    stack.put(new Token<Character>(char.class), getCharAdapter());
    stack.put(new Token<Character>(Character.class), getCharAdapter());

    stack.put(new Token<String>(String.class), getStringAdapter());
    stack.put(new Token<CharSequence>(CharSequence.class), getCharSequenceAdapter());
    stack.put(new Token<StringBuilder>(StringBuilder.class), getStringBuilderAdapter());
    stack.put(new Token<StringBuffer>(StringBuffer.class), getStringBufferAdapter());
  }

  /**
   * Adds Date and Calendar Adapters.
   *
   * @param stack
   */
  private void includeTimeAdapters(Map<Token<?>, JsonAdapter<?>> stack) {
    stack.put(new Token<Date>(Date.class), getDateAdapter());
    stack.put(new Token<Calendar>(Calendar.class), getCalendarAdapter());
  }

  /**
   * Adds Boolean adapters.
   *
   * @param stack
   */
  private void includeBoolAdapters(Map<Token<?>, JsonAdapter<?>> stack) {
    stack.put(new Token<Boolean>(boolean.class), getBooleanAdapter());
    stack.put(new Token<Boolean>(Boolean.class), getBooleanAdapter());
  }

  /**
   * @param map
   * @param stack
   */
  private void embedAdapters(Map<Type, JsonAdapter<?>> map, Map<Token<?>, JsonAdapter<?>> stack) {
    for (Map.Entry<Type, JsonAdapter<?>> entry : map.entrySet()) {
      Type key = entry.getKey();
      JsonAdapter<?> value = entry.getValue();

      stack.put(new Token<>(key), value);
    }
  }

  /**
   * @param map
   * @param stack
   * @param serializerStack
   */
  private void embedSerializers(Map<Object, JsonSerializer<?>> map,
      Map<Token<?>, JsonAdapter<?>> stack,
      Map<String, JsonRuntimeSerializer<?>> serializerStack) {

    for (Map.Entry<Object, JsonSerializer<?>> entry : map.entrySet()) {
      Object key = entry.getKey();
      JsonSerializer<?> value = entry.getValue();

      if (key instanceof Type) {
        Type type = (Type) key;
        Token<?> token = new Token<>(type);
        stack.put(token, mergeWithDeserializer(token, value));
      } else if (key instanceof String) {
        String typeString = (String) key;
        serializerStack.put(typeString, new JsonRuntimeSerializer<>(value));
      }
    }

  }

  /**
   * <p>
   * Transforms given serialize to an adapter.
   * </p>
   *
   * @param <T>
   * @param token
   * @param serializer
   * @return
   */
  private <T> JsonAdapter<T> mergeWithDeserializer(final Token<?> token, final JsonSerializer<T> serializer) {
    final Json json = this;
    return new JsonAdapter<T>() {
      @Override
      public void write(JsonBuffer buffer, T object) {
        serializer.write(buffer, new JsonRuntimeSerializerContext<>(json, object));
      }

      @Override
      public T read(JsonStream stream) {
        JsonAdapter<T> adapter = (JsonAdapter<T>) getAdapter(token);
        return adapter.read(stream);
      }
    };
  }

  /**
   * @param map
   * @param stack
   * @param deserializerStack
   */
  private void embedDeserializers(Map<Object, JsonDeserializer<?>> map,
      Map<Token<?>, JsonAdapter<?>> stack,
      Map<String, JsonRuntimeDeserializer<?>> deserializerStack) {

    for (Map.Entry<Object, JsonDeserializer<?>> entry : map.entrySet()) {
      Object key = entry.getKey();
      JsonDeserializer<?> value = entry.getValue();

      if (key instanceof Type) {
        Type type = (Type) key;
        Token<?> token = new Token<>(type);
        stack.put(token, mergeWithSerializer(token, value));
      } else if (key instanceof String) {
        String typeString = (String) key;
        deserializerStack.put(typeString, new JsonRuntimeDeserializer<>(value));
      }
    }

  }

  /**
   * <p>
   * Transforms given deserializer to an adapter.
   * </p>
   *
   * @param <T>
   * @param token
   * @param deserializer
   * @return
   */
  private <T> JsonAdapter<T> mergeWithSerializer(final Token<?> token, final JsonDeserializer<T> deserializer) {
    final Json json = this;
    return new JsonAdapter<T>() {
      @Override
      public void write(JsonBuffer buffer, T object) {
        JsonAdapter<T> adapter = (JsonAdapter<T>) getAdapter(token);
        adapter.write(buffer, object);
      }

      @Override
      public T read(JsonStream stream) {
        return deserializer.read(stream, new JsonRuntimeDeserializerContext<T>(json));
      }
    };
  }

  /**
   * @see tr.com.terrayazilim.json.JsonNull for definition of {@code null}.
   *
   * @return JsonNullAdapter
   */
  private JsonAdapter getNullAdapter() {
    return new JsonAdapter() {
      @Override
      public void write(JsonBuffer buffer, Object object) {
        buffer.value(new JsonNull().toString());
      }

      @Override
      public Object read(JsonStream stream) {
        return null;
      }
    };
  }

  /**
   * <p>
   * Uses {@code SimpleDateFormat}. Default format for both serialization and deserialization is
   * {@literal dd-MM-yyy HH:mm:ss}.
   * </p>
   *
   * @return {@link java.util.Date} serialization and deserialization adapter.
   */
  private JsonAdapter<Date> getDateAdapter() {
    return new JsonAdapter<Date>() {
      final String pattern = "dd-MM-yyy HH:mm:ss";

      @Override
      public void write(JsonBuffer buffer, Date object) {
        SimpleDateFormat simpleDateFormat;
        if (policy.getDateFormat() == null) {
          simpleDateFormat = new SimpleDateFormat(pattern);
        } else {
          simpleDateFormat = new SimpleDateFormat(policy.getDateFormat());
        }

        String date = simpleDateFormat.format(object);
        buffer.value(date);
      }

      @Override
      public Date read(JsonStream stream) {
        SimpleDateFormat sdf;
        if (policy.getDateFormat() == null) {
          sdf = new SimpleDateFormat(pattern);
        } else {
          sdf = new SimpleDateFormat(policy.getDateFormat());
        }

        Date date = null;
        try {
          date = sdf.parse(stream.nextString());
        } catch (ParseException ex) {
        }

        return date;
      }
    };
  }

  /**
   * <p>
   * Uses {@code SimpleDateFormat}. Default format for both serialization and deserialization is
   * {@literal dd-MM-yyy HH:mm:ss}.
   * </p>
   *
   * @return {@link java.util.Calendar} serialization and deserialization adapter.
   */
  private JsonAdapter<Calendar> getCalendarAdapter() {
    return new JsonAdapter<Calendar>() {
      final String pattern = "dd-MM-yyy HH:mm:ss";

      @Override
      public void write(JsonBuffer buffer, Calendar object) {
        SimpleDateFormat dateFormat = new SimpleDateFormat(pattern);
        String value = dateFormat.format(object.getTime());

        buffer.value(value);
      }

      @Override
      public Calendar read(JsonStream stream) {
        SimpleDateFormat sdf;
        if (policy.getDateFormat() == null) {
          sdf = new SimpleDateFormat(pattern);
        } else {
          sdf = new SimpleDateFormat(policy.getDateFormat());
        }

        Calendar calendar = Calendar.getInstance();
        String value = stream.nextString();
        Date date = null;
        try {
          date = sdf.parse(value);
        } catch (ParseException ex) {
        }

        calendar.setTime(date);
        return calendar;
      }
    };
  }

  private JsonAdapter<JsonArray> getJsonArrayAdapter() {
    return new JsonAdapterFactory() {
      @Override
      public <T> JsonAdapter<T> call(final Json json, Token<T> token) {
        return (JsonAdapter<T>) new JsonAdapter<JsonArray>() {
          @Override
          public void write(JsonBuffer buffer, JsonArray object) {
            buffer.beginArray();

            Iterator<Object> it = object.iterator();
            while (it.hasNext()) {
              Object next = it.next();
              JsonAdapter adapter = json.getAdapter(new Token<>(next.getClass()));
              adapter.write(buffer, next);
            }

            buffer.endArray();
          }

          @Override
          public JsonArray read(JsonStream stream) {
            return (JsonArray) stream.value();
          }
        };
      }
    }.call(this, new Token<JsonArray>(JsonArray.class));
  }

  private JsonAdapter<JsonObject> getJsonObjectAdapter() {
    return new JsonAdapterFactory() {
      @Override
      public <T> JsonAdapter<T> call(final Json json, Token<T> token) {
        return (JsonAdapter<T>) new JsonAdapter<JsonObject>() {
          @Override
          public void write(JsonBuffer buffer, JsonObject object) {
            Set<String> keySet = object.keySet();
            buffer.beginObject();

            for (String eachKey : keySet) {
              Object value = object.get(eachKey);
              JsonAdapter adapter = json.getAdapter(new Token<>(value.getClass()));
              buffer.key(eachKey);
              adapter.write(buffer, value);
            }

            buffer.endObject();
          }

          @Override
          public JsonObject read(JsonStream stream) {
            return (JsonObject) stream.value();
          }
        };
      }
    }.call(this, new Token<JsonObject>(JsonObject.class));
  }

  private JsonAdapter<JsonNull> getJsonNullAdapter() {
    return new JsonAdapter<JsonNull>() {
      @Override
      public void write(JsonBuffer buffer, JsonNull object) {
        buffer.value(object.toString());
      }

      @Override
      public JsonNull read(JsonStream stream) {
        return (JsonNull) stream.value();
      }
    };
  }

  private JsonAdapter<JsonLiteral> getJsonLiteralAdapter() {
    return new JsonAdapter<JsonLiteral>() {
      @Override
      public void write(JsonBuffer buffer, JsonLiteral object) {
        Object literal = object.getValue();
        if (literal == null) {
          buffer.value(new JsonNull().toString());
        } else {
          buffer.value(literal);
        }
      }

      @Override
      public JsonLiteral read(JsonStream stream) {
        return (JsonLiteral) stream.value();
      }
    };
  }

  /**
   * {@link java.lang.Enum} adapter factory.
   */
  private <T extends Enum<T>> JsonAdapterFactory getEnumAdapterFactory() {
    return new JsonAdapterFactory() {
      @Override
      public <E> JsonAdapter<E> call(Json json, final Token<E> token) {
        return (JsonAdapter<E>) new JsonAdapter<T>() {
          @Override
          public void write(JsonBuffer buffer, T object) {
            buffer.value(object.toString());
          }

          @Override
          public T read(JsonStream stream) {
            String enumId = stream.nextString();

            Class<?> declaringClass = token.getRawType();
            for (Object eachEnum : declaringClass.getEnumConstants()) {
              Enum enumType = (Enum) eachEnum;
              if (enumType.name().equals(enumId)) {
                return (T) enumType;
              }
            }

            return null;
          }
        };
      }
    };
  }

  /**
   * @return An JsonAdapter for both serialization and deserialization of {@code int} and
   * {@link java.lang.Integer}
   */
  private JsonAdapter<Integer> getIntAdapter() {
    return new JsonAdapter<Integer>() {
      @Override
      public void write(JsonBuffer buffer, Integer object) {
        buffer.value(object);
      }

      @Override
      public Integer read(JsonStream stream) {
        return stream.nextInt();
      }
    };
  }

  /**
   * @return An JsonAdapter for both serialization and deserialization of {@code float} and
   * {@link java.lang.Float}
   */
  private JsonAdapter<Float> getFloatAdapter() {
    return new JsonAdapter<Float>() {
      @Override
      public void write(JsonBuffer buffer, Float object) {
        buffer.value(object);
      }

      @Override
      public Float read(JsonStream stream) {
        return stream.nextFloat();
      }
    };
  }

  /**
   * @return An JsonAdapter for both serialization and deserialization of {@code double} and
   * {@link java.lang.Double}
   */
  private JsonAdapter<Double> getDoubleadapter() {
    return new JsonAdapter<Double>() {
      @Override
      public void write(JsonBuffer buffer, Double object) {
        buffer.value(object);
      }

      @Override
      public Double read(JsonStream stream) {
        return stream.nextDouble();
      }
    };
  }

  /**
   * @return An JsonAdapter for both serialization and deserialization of {@code long} and
   * {@link java.lang.Long}
   */
  private JsonAdapter<Long> getLongAdapter() {
    return new JsonAdapter<Long>() {
      @Override
      public void write(JsonBuffer buffer, Long object) {
        buffer.value(object);
      }

      @Override
      public Long read(JsonStream stream) {
        return stream.nextLong();
      }
    };
  }

  /**
   * @return An JsonAdapter for both serialization and deserialization of {@code short} and
   * {@link java.lang.Short}
   */
  private JsonAdapter<Short> getShortAdapter() {
    return new JsonAdapter<Short>() {
      @Override
      public void write(JsonBuffer buffer, Short object) {
        buffer.value(object);
      }

      @Override
      public Short read(JsonStream stream) {
        return stream.nextShort();
      }
    };
  }

  /**
   * @return An JsonAdapter for both serialization and deserialization of {@code byte} and
   * {@link java.lang.Byte}
   */
  private JsonAdapter<Byte> getByteAdapter() {
    return new JsonAdapter<Byte>() {
      @Override
      public void write(JsonBuffer buffer, Byte object) {
        buffer.value(object);
      }

      @Override
      public Byte read(JsonStream stream) {
        return stream.nextByte();
      }
    };
  }

  /**
   * @return An JsonAdapter for both serialization and deserialization of {@code char} and
   * {@link java.lang.Character}
   */
  private JsonAdapter<Character> getCharAdapter() {
    return new JsonAdapter<Character>() {
      @Override
      public void write(JsonBuffer buffer, Character object) {
        buffer.value(object);
      }

      @Override
      public Character read(JsonStream stream) {
        return stream.nextChar();
      }
    };
  }

  /**
   * @return An JsonAdapter for both serialization and deserialization of {@link java.lang.String}
   */
  private JsonAdapter<String> getStringAdapter() {
    return new JsonAdapter<String>() {
      @Override
      public void write(JsonBuffer buffer, String object) {
        buffer.value(object);
      }

      @Override
      public String read(JsonStream stream) {
        return stream.nextString();
      }
    };
  }

  /**
   * @return An JsonAdapter for both serialization and deserialization of
   * {@link java.lang.CharSequence}
   */
  private JsonAdapter<CharSequence> getCharSequenceAdapter() {
    return new JsonAdapter<CharSequence>() {
      @Override
      public void write(JsonBuffer buffer, CharSequence object) {
        buffer.value(object);
      }

      @Override
      public CharSequence read(JsonStream stream) {
        return stream.nextString();
      }
    };
  }

  /**
   * @return An JsonAdapter for both serialization and deserialization of
   * {@link java.lang.StringBuilder}
   */
  private JsonAdapter<StringBuilder> getStringBuilderAdapter() {
    return new JsonAdapter<StringBuilder>() {
      @Override
      public void write(JsonBuffer buffer, StringBuilder object) {
        buffer.value(object.toString());
      }

      @Override
      public StringBuilder read(JsonStream stream) {
        return new StringBuilder(stream.nextString());
      }
    };
  }

  /**
   * @return An JsonAdapter for both serialization and deserialization of
   * {@link java.lang.StringBuffer}
   */
  private JsonAdapter<StringBuffer> getStringBufferAdapter() {
    return new JsonAdapter<StringBuffer>() {
      @Override
      public void write(JsonBuffer buffer, StringBuffer object) {
        buffer.value(object.toString());
      }

      @Override
      public StringBuffer read(JsonStream stream) {
        return new StringBuffer(stream.nextString());
      }
    };
  }

  /**
   * @return An JsonAdapter for both serialization and deserialization of {@code boolean} and
   * {@link java.lang.Boolean}
   */
  private JsonAdapter<Boolean> getBooleanAdapter() {
    return new JsonAdapter<Boolean>() {
      @Override
      public void write(JsonBuffer buffer, Boolean object) {
        buffer.value(object);
      }

      @Override
      public Boolean read(JsonStream stream) {
        return stream.nextBool();
      }
    };
  }

  /**
   * @return An JsonAdapter for both serialization and deserialization of
   * {@link java.math.BigDecimal}
   */
  private JsonAdapter<BigDecimal> getBigDecimalAdapter() {
    return new JsonAdapter<BigDecimal>() {
      @Override
      public void write(JsonBuffer buffer, BigDecimal object) {
        buffer.value(object);
      }

      @Override
      public BigDecimal read(JsonStream stream) {
        return new BigDecimal(stream.nextString());
      }
    };
  }

  /**
   * @return An JsonAdapter for both serialization and deserialization of {@link java.util.UUID}
   */
  private JsonAdapter<UUID> getUUIDAdapter() {
    return new JsonAdapter<UUID>() {
      @Override
      public void write(JsonBuffer buffer, UUID object) {
        buffer.value(object.toString());
      }

      @Override
      public UUID read(JsonStream stream) {
        throw new UnsupportedOperationException("Not supported yet.");
      }
    };
  }

  /**
   * <p>
   * Primal {@code JsonAdapterFactory} for parsing. It processes the {@code JsonGraph} structure.
   * </p>
   *
   * @param <T>
   * @return
   */
  private <T> JsonAdapterFactory getGraphAdapterFactory() {
    return new JsonAdapterFactory() {
      final String getBufferKey(String key) {
        if (policy.getNamingPolicy() == null) {
          return key;
        }

        switch (policy.getNamingPolicy()) {
          case DEFAULT:
            return key;
          case LOWERCASE:
            return key.toLowerCase(Locale.ENGLISH);
          case UPPERCASE:
            return key.toUpperCase(Locale.ENGLISH);
        }

        return key;
      }

      @Override
      public <T> JsonAdapter<T> call(final Json json, final Token<T> token) {
        return new JsonAdapter<T>() {
          @Override
          public void write(JsonBuffer buffer, T instance) {
            JsonGraph<T> graph = new JsonGraph<>();
            graph.buildWriterGraph(json, instance, token);
            Queue<JsonGraph.Element<?>> stack = graph.getQueue();

            buffer.beginObject();
            while (!stack.isEmpty()) {
              final JsonGraph.Element<?> element = stack.poll();
              final String key = element.getKeyToken().getName();
              final Object value = element.getKeyToken().getValue(instance);
              buffer.key(getBufferKey(key));

              JsonSerializer serializer = element.getSerializer();
              if (serializer != null) {
                serializer.write(buffer, new JsonRuntimeSerializerContext(json, value));
              } else {
                JsonAdapter adapter = element.getAdapter();
                adapter.write(buffer, value);
              }
            }

            buffer.endObject();
          }

          @Override
          public T read(JsonStream stream) {
            ConstructorToken<T> constructorToken = new ConstructorToken<>(token.getRawType());
            T instance = constructorToken.getDefaultConstructor();
            if (instance == null) {
              throw new JsonException("Object instance doesn't have default constructor. "
                  + "{Hint: ErrorCode: 107}");
            }

            JsonGraph<T> graph = new JsonGraph<>();
            graph.buildReaderGraph(json, stream, token);
            Queue<JsonGraph.Element<?>> stack = graph.getQueue();
            while (!stack.isEmpty()) {
              final JsonGraph.Element<?> element = stack.poll();
              final FieldToken keyToken = element.getKeyToken();
              final JsonStream broadcast = element.getStream();

              Object value;
              JsonDeserializer deserializer = element.getDeserializer();
              if (deserializer != null) {
                value = deserializer.read(broadcast, new JsonRuntimeDeserializerContext(json));
              } else {
                JsonAdapter adapter = element.getAdapter();
                value = adapter.read(broadcast);
              }

              Field field = keyToken.getField();
              field.setAccessible(true);
              try {
                field.set(instance, value);
              } catch (IllegalArgumentException | IllegalAccessException ex) {
              }
            }

            return instance;
          }
        };
      }
    };
  }

  /**
   * For avoiding {@link NullPointerException} in 
   *  {@link #getAdapter(tr.com.terrayazilim.json.bind.Token) }
   */
  private final Token<Object> NULL_TOKEN = new Token<Object>() {
  };

  /**
   * @param <T>
   * @param key
   * @return
   */
  public <T> JsonDeserializer<T> getDeserializer(String key) {
    return (JsonDeserializer<T>) deserializers.get(key);
  }

  /**
   * @param <T>
   * @param key
   * @return
   */
  public <T> JsonSerializer<T> getSerializer(String key) {
    return (JsonSerializer<T>) serializers.get(key);
  }

  /**
   * <p>
   * Determines adapter in runtime. It is thread safe and supports caching.
   * </p>
   *
   * @param <T>
   * @param typeToken
   * @return
   */
  public <T> JsonAdapter<T> getAdapter(final Token<T> typeToken) {
    Token<?> token = Objects.nonNull(typeToken) ? typeToken : NULL_TOKEN;
    JsonAdapter<?> adapter = getCache().get(token);
    if (adapter != null) {
      return (JsonAdapter<T>) adapter;
    }

    Map<Token<?>, JsonRuntimeAdapter<?>> safeCall = callback().get();
    boolean clear = false;
    if (safeCall == null) {
      safeCall = new HashMap<>();
      callback().set(safeCall);
      clear = true;
    }

    JsonRuntimeAdapter<T> next = (JsonRuntimeAdapter<T>) safeCall.get(token);
    if (next != null) {
      return next;
    }

    try {
      JsonRuntimeAdapter<T> call = new JsonRuntimeAdapter<>();
      safeCall.put(token, call);

      JsonAdapter<T> ref = (JsonAdapter<T>) getProxy().get(token);
      if (ref != null) {
        getCache().put(token, ref);
        return ref;
      } else {
        if (token.isArray()) {
          ref = (JsonAdapter<T>) new JsonArrayAdapter<>(this, token);
        } else if (token.isCollection()) {
          ref = (JsonAdapter<T>) new JsonCollectionAdapter<>(this, token);
        } else if (token.isMap()) {
          ref = new JsonMapAdapter(this, token);
        } else if (token.isEnum()) {
          ref = (JsonAdapter<T>) getEnumAdapterFactory().call(this, token);
        } else {
          ref = (JsonAdapter<T>) getGraphAdapterFactory().call(this, token);
        }

        if (ref != null) {
          getCache().put(token, ref);
          return ref;
        }
      }

      throw new JsonException("Type cannot handle. {Hint: ErrorCode: 106}");
    } finally {
      safeCall.remove(token);

      if (clear) {
        callback().remove();
      }
    }
  }

  /**
   * @param <T>
   * @param instance
   * @return
   */
  public <T> JsonObject toJson(T instance) {
    if (instance == null) {
      return new JsonObject();
    }

    Token<T> token = new Token<>(instance.getClass());

    JsonAdapterFactory factory = getGraphAdapterFactory();
    JsonAdapter<T> adapter = factory.call(this, token);

    JsonBuffer buffer = new JsonBuffer();
    adapter.write(buffer, instance);

    return new JsonObject(buffer.toString());
  }

  /**
   * <p>
   * Use it for instances that has generic signature. </p>
   *
   * @param <T>
   * @param instance
   * @param token
   * @return
   */
  public <T> String toJson(T instance, Token<T> token) {
    if (instance == null) {
      return "";
    }

    JsonAdapterFactory factory = getGraphAdapterFactory();
    JsonAdapter<T> adapter = factory.call(this, token);

    JsonBuffer buffer = new JsonBuffer();
    adapter.write(buffer, instance);

    return buffer.toString();
  }

  /**
   * @param <R>
   * @param instance
   * @param indent
   * @return
   */
  public <R> String toJson(R instance, int indent) {
    return toJson(instance).toString(indent);
  }

  /**
   * @param <R>
   * @param json
   * @param token
   * @return
   */
  public <R> R fromJson(String json, Token<R> token) {
    if (!isJsonObject(json)) {
      throw new JsonException("Given String is not Acceptable JSON. {Hint: ErrorCode:105}");
    } else if (Strings.nullOrBlank(json)) {
      return null;
    }

    JsonAdapterFactory factory = getGraphAdapterFactory();
    JsonAdapter<R> adapter = factory.call(this, token);

    JsonStream stream = new JsonStream(new JsonObject(json));
    R instance = adapter.read(stream);

    return instance;
  }

  /**
   * @param <T>
   * @param jsonObject
   * @param token
   * @return
   */
  public <T> T fromJson(JsonObject jsonObject, Token<T> token) {
    return fromJson(jsonObject.toString(), token);
  }

  /**
   * @return
   */
  private Map<Token<?>, JsonAdapter<?>> getProxy() {
    return this.proxy;
  }

  /**
   * @return
   */
  private Map<Token<?>, JsonAdapter<?>> getCache() {
    return this.cache;
  }

  /**
   * It returns copy of {@code JsonPolicy}, not the original one.
   *
   * @return For sake of Immutablity, it returns copy of {@code JsonPolicy}.
   */
  public JsonPolicy getPolicy() {
    return new JsonPolicy(policy);
  }

  /**
   * @return
   */
  private ThreadLocal<Map<Token<?>, JsonRuntimeAdapter<?>>> callback() {
    return this.adapters;
  }

  /**
   * @param json
   * @return
   */
  @SuppressWarnings("ResultOfObjectAllocationIgnored")
  public static boolean isJson(final String json) {
    try {
      new JsonObject(json);
    } catch (JsonException ex) {
      try {
        new JsonArray(json);
      } catch (JsonException ex1) {
        return false;
      }
    }

    return true;
  }

  /**
   * @param json
   * @return
   */
  @SuppressWarnings("ResultOfObjectAllocationIgnored")
  public static boolean isJsonObject(final String json) {
    try {
      new JsonObject(json);
    } catch (Exception e) {
      return false;
    }

    return true;
  }

  /**
   * @param jsonArray
   * @return
   */
  @SuppressWarnings("ResultOfObjectAllocationIgnored")
  public static boolean isJsonArray(final String jsonArray) {
    try {
      new JsonArray(jsonArray);
    } catch (Exception e) {
      return false;
    }

    return true;
  }

  /**
   * @version 1.0.3
   * @since 1.0.3
   * @param <T>
   */
  static class JsonRuntimeDeserializerContext<T> implements JsonDeserializerContext<T> {

    private final Json json;

    public JsonRuntimeDeserializerContext(final Json json) {
      this.json = json;
    }

    @Override
    public Json getJson() {
      return this.json;
    }

  }

  /**
   * @version 1.0.3
   * @since 1.0.3
   * @param <T>
   */
  static class JsonRuntimeSerializerContext<T> implements JsonSerializerContext<T> {

    private final Json json;
    private final T target;

    public JsonRuntimeSerializerContext(final Json json, final T target) {
      this.json = json;
      this.target = target;
    }

    @Override
    public Json getJson() {
      return this.json;
    }

    @Override
    public T getTarget() {
      return this.target;
    }

  }

  /**
   * @version 1.0.3
   * @since 1.0.3
   * @param <T>
   */
  static class JsonRuntimeDeserializer<T> extends JsonDeserializer<T> {

    private final JsonDeserializer<T> runtimeDeserializer;

    public JsonRuntimeDeserializer(JsonDeserializer<T> deserializer) {
      this.runtimeDeserializer = deserializer;
    }

    @Override
    public T read(JsonStream stream, JsonDeserializerContext<T> context) {
      if (runtimeDeserializer == null) {
        throw new JsonException("JsonRuntimeDeserializer doesn't have any "
            + "JsonDeserializer to proceed deserialization. {Hint: ErrorCode:104}");
      }

      return runtimeDeserializer.read(stream, context);
    }
  }

  /**
   * @version 1.0.3
   * @since 1.0.3
   * @param <T>
   */
  static class JsonRuntimeSerializer<T> extends JsonSerializer<T> {

    private final JsonSerializer<T> runtimeSerializer;

    public JsonRuntimeSerializer(JsonSerializer<T> serializer) {
      this.runtimeSerializer = serializer;
    }

    @Override
    public void write(JsonBuffer buffer, JsonSerializerContext<T> context) {
      if (runtimeSerializer == null) {
        throw new JsonException("JsonRuntimeSerializer doesn't have any "
            + "JsonSerializer to proceed serialization. {Hint: ErrorCode:103}");
      }

      runtimeSerializer.write(buffer, context);
    }
  }

  /**
   * @version 1.0.3
   * @since 1.0.3
   * @param <T>
   */
  static class JsonRuntimeAdapter<T> extends JsonAdapter<T> {

    private JsonAdapter<T> runtimeAdapter;

    /**
     * @param runtimeAdapter
     */
    public void setRuntimeAdapter(JsonAdapter<T> runtimeAdapter) {
      this.runtimeAdapter = runtimeAdapter;
    }

    @Override
    public void write(JsonBuffer buffer, T object) {
      if (runtimeAdapter == null) {
        throw new JsonException("JsonRuntimeAdapter doesn't have any "
            + "JsonAdapter to proceed serialization. {Hint: ErrorCode:101}");
      }

      runtimeAdapter.write(buffer, object);
    }

    @Override
    public T read(JsonStream stream) {
      if (runtimeAdapter == null) {
        throw new JsonException("JsonRuntimeAdapter doesn't have any "
            + "JsonAdapter to proceed deserialization. {Hint: ErrorCode:102}");
      }

      return runtimeAdapter.read(stream);
    }
  }

}
