/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * This file is part of terrayazilim-json project.
 *
 * This file incorporates work covered by
 * the following copyright and permission notices:
 *
 * Copyright (C) 2018 Terra Yazılım Bilişim Hiz. Elek. Dan. Oto. ve Loj. Tic. Ltd. Şti. | info [at] terrayazilim [dot] com [dot] tr
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package tr.com.terrayazilim.json;

import java.lang.reflect.Type;
import java.util.Map;
import java.util.Set;
import tr.com.terrayazilim.core.Objects;
import tr.com.terrayazilim.json.JsonPolicy.KeyNamingPolicy;

/**
 * @author M.Çağrı Tepebaşılı - cagritepebasili [at] protonmail [dot] com
 * @version 1.0.3
 * @since 1.0.3
 */
public final class JsonBuilder {

  private JsonPolicy policy;

  public JsonBuilder() {
  }
  
  /**
   * @return 
   */
  public Json toJson() {
    return new Json(getPolicy());
  }
  
  /**
   * @return 
   */
  public JsonBuilder enableNullSerialization() {
    getPolicy().setNullSerialization(true);
    return this;
  }
  
  public JsonBuilder disableNullSerialization() {
    getPolicy().setNullSerialization(false);
    return this;
  }
  
  public JsonBuilder enableNullDeserialization() {
    getPolicy().setNullDeserialization(Boolean.TRUE);
    return this;
  }
  
  public JsonBuilder disableNullDeserialization() {
    getPolicy().setNullDeserialization(Boolean.FALSE);
    return this;
  }
  
  /**
   * @param keyNamingPolicy
   * @return
   */
  public JsonBuilder setKeyNamingPolicy(JsonPolicy.KeyNamingPolicy keyNamingPolicy) {
    getPolicy().setNamingPolicy(keyNamingPolicy);
    return this;
  }
  
  /**
   * @param dateFormat
   * @return
   */
  public JsonBuilder setDateFormat(String dateFormat) {
    getPolicy().setDateFormat(dateFormat);
    return this;
  }
  
  /**
   * @return
   */
  public JsonBuilder enableEnumSerialization() {
    getPolicy().setEnumSerialization(Boolean.TRUE);
    return this;
  }
  
  public JsonBuilder disableEnumSerialization() {
    getPolicy().setEnumSerialization(Boolean.FALSE);
    return this;
  }
  
  /**
   * @return
   */
  public JsonBuilder enableEnumDeserialization() {
    getPolicy().setEnumDeserialization(Boolean.TRUE);
    return this;
  }
  
  public JsonBuilder disableEnumDeserialization(Boolean bool) {
    getPolicy().setEnumDeserialization(Boolean.FALSE);
    return this;
  }
  
  /**
   * @return
   */
  public JsonBuilder enableSuperClassSerialization() {
    getPolicy().setSuperClassSerialization(Boolean.TRUE);
    return this;
  }
  
  public JsonBuilder disableSuperClassSerialization() {
    getPolicy().setSuperClassSerialization(Boolean.FALSE);
    return this;
  }
  
  /**
   * @return
   */
  public JsonBuilder enableSuperClassDeserialization() {
    getPolicy().setSuperClassDeserialization(Boolean.TRUE);
    return this;
  }
  
  public JsonBuilder disableSuperClassDeserialization() {
    getPolicy().setSuperClassDeserialization(Boolean.FALSE);
    return this;
  }
  
  /**
   * @param ignoreType
   * @return 
   */
  public JsonBuilder ignore(Type ignoreType) {
    getPolicy().getIgnoredTypes().add(ignoreType);
    return this;
  }
  
  /**
   * @param types
   * @return 
   */
  public JsonBuilder registerIgnoredTypes(Set<Type> types) {
    getPolicy().setIgnoredTypes(types);
    return this;
  }
  
  /**
   * @param ignoreField
   * @return 
   */
  public JsonBuilder ignore(String ignoreField) {
    getPolicy().getIgnoredFields().add(ignoreField);
    return this;
  }
  
  /**
   * After called this method, fields modifiers in {@link JsonPolicy} become redundant.
   * 
   * @return {@code JsonBuilder}
   */
  public JsonBuilder disableFieldModifierChecking() {
    getPolicy().setIgnoreFieldModifiers(Boolean.TRUE);
    return this;
  }

  /**
   * @return 
   */
  public JsonBuilder enableFieldModifierChecking() {
    getPolicy().setIgnoreFieldModifiers(Boolean.FALSE);
    return this;
  }
  
  /**
   * @param fields
   * @return 
   */
  public JsonBuilder registerIgnoredFields(Set<String> fields) {
    getPolicy().setIgnoredFields(fields);
    return this;
  }
  
  /**
   * @param <R>
   * @param type
   * @param adapter
   * @return 
   */
  public <R> JsonBuilder registerAdapter(Type type, JsonAdapter<R> adapter) {
    getPolicy().getAdapters().put(type, adapter);
    return this;
  }
  
  /**
   * @param <R>
   * @param adapters
   * @return 
   */
  public <R> JsonBuilder setAdapters(Map<Type, JsonAdapter<?>> adapters) {
    getPolicy().setAdapters(adapters);
    return this;
  }
  
  /**
   * @param <R>
   * @param fieldName
   * @param serializer 
   * @return
   */
  public <R> JsonBuilder registerSerializer(String fieldName, JsonSerializer<R> serializer) {
    getPolicy().getSerializers().put(fieldName, serializer);
    return this;
  }
  
  /**
   * @param <R>
   * @param type
   * @param serializer
   * @return 
   */
  public <R> JsonBuilder registerSerializer(Type type, JsonSerializer<R> serializer) {
    getPolicy().getSerializers().put(type, serializer);
    return this;
  }
  
  /**
   * @param <R>
   * @param fieldName
   * @param deserializer
   * @return
   */
  public <R> JsonBuilder registerDeserializer(String fieldName, JsonDeserializer<R> deserializer) {
    getPolicy().getDeserializers().put(fieldName, deserializer);
    return this;
  }
  
  /**
   * @param <R>
   * @param type
   * @param deserializer
   * @return 
   */
  public <R> JsonBuilder registerDeserializer(Type type, JsonDeserializer<R> deserializer) {
    getPolicy().getDeserializers().put(type, deserializer);
    return this;
  }
  
  /**
   * @param codecs
   * @return
   */
  public JsonBuilder setSerializers(Map<Object, JsonSerializer<?>> codecs) {
    getPolicy().setSerializer(codecs);
    return this;
  }
  
  /**
   * @param codecs
   * @return
   */
  public JsonBuilder setDeserializers(Map<Object, JsonDeserializer<?>> codecs) {
    getPolicy().setDeserializers(codecs);
    return this;
  }
  
  /**
   * @param namingPolicy
   * @return
   */
  public JsonBuilder registerKeyNamingPolicy(KeyNamingPolicy namingPolicy) {
    getPolicy().setNamingPolicy(namingPolicy);
    return this;
  }
  
  /**
   * @param modifier
   * @return
   */
  public JsonBuilder registerAllowedFieldModifier(Integer modifier) {
    getPolicy().getAllowedFieldModifiers().add(modifier);
    return this;
  }
  
  /**
   * @param mods
   * @return
   */
  public JsonBuilder registerAllowedFieldModifiers(Set<Integer> mods) {
    getPolicy().setAllowedFieldModifiers(mods);
    return this;
  }
  
  /**
   * @param modifier
   * @return
   */
  public JsonBuilder registerAllowedSubClassFieldModifier(Integer modifier) {
    getPolicy().getAllowedSubClassFieldModifiers().add(modifier);
    return this;
  }
  
  /**
   * @param mods
   * @return
   */
  public JsonBuilder registerAllowedSubClassFieldModifiers(Set<Integer> mods) {
    getPolicy().setAllowedSubClassFieldModifiers(mods);
    return this;
  }
  
  /**
   * @return
   */
  public JsonPolicy getPolicy() {
    if (Objects.isNull(this.policy)) {
      this.policy = new JsonPolicy();
    }
    
    return this.policy;
  }
}
