/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * This file is part of terrayazilim-json project.
 *
 * This file incorporates work covered by
 * the following copyright and permission notices:
 *
 * Copyright (C) 2018 Terra Yazılım Bilişim Hiz. Elek. Dan. Oto. ve Loj. Tic. Ltd. Şti. | info [at] terrayazilim [dot] com [dot] tr
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package tr.com.terrayazilim.json;

import java.lang.reflect.Type;
import java.lang.reflect.WildcardType;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Locale;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import tr.com.terrayazilim.core.Ints;
import tr.com.terrayazilim.core.Objects;
import tr.com.terrayazilim.core.collection.Sets;
import tr.com.terrayazilim.json.annotation.Adapter;
import tr.com.terrayazilim.json.annotation.Deserializer;
import tr.com.terrayazilim.json.annotation.Ignored;
import tr.com.terrayazilim.json.annotation.Serializer;
import tr.com.terrayazilim.json.stream.JsonStream;
import tr.com.terrayazilim.json.type.ConstructorToken;
import tr.com.terrayazilim.json.type.FieldToken;
import tr.com.terrayazilim.json.type.Token;

/**
 * @author M.Çağrı Tepebaşılı - cagritepebasili [at] protonmail [dot] com
 * @version 1.0.3
 * @since 1.0.3
 * @param <T>
 */
public class JsonGraph<T> {

  /**
   * @version 1.0.3
   * @since 1.0.3
   * @param <T>
   */
  static class Element<T> {

    private final JsonDeserializer<T> deserializer;
    private final JsonStream stream;

    private final JsonAdapter<T> adapter;

    private final JsonSerializer<T> serializer;
    private final FieldToken keyToken;

    public Element(FieldToken token, JsonStream stream, JsonAdapter<T> adapter) {
      this.deserializer = null;
      this.stream = stream;
      this.adapter = adapter;
      this.serializer = null;
      this.keyToken = token;
    }

    public Element(FieldToken token, JsonStream stream, JsonDeserializer<T> deserializer) {
      this.deserializer = deserializer;
      this.stream = stream;
      this.adapter = null;
      this.serializer = null;
      this.keyToken = token;
    }

    public Element(FieldToken keyToken, JsonAdapter<T> adapter, JsonSerializer<T> serializer) {
      this.serializer = serializer;
      this.adapter = adapter;
      this.keyToken = keyToken;
      this.deserializer = null;
      this.stream = null;
    }

    public Element(FieldToken keyToken, JsonAdapter<T> adapter) {
      this.keyToken = keyToken;
      this.adapter = adapter;
      this.serializer = null;
      this.deserializer = null;
      this.stream = null;
    }

    public JsonAdapter<T> getAdapter() {
      return adapter;
    }

    public FieldToken getKeyToken() {
      return keyToken;
    }

    public JsonSerializer<T> getSerializer() {
      return serializer;
    }

    public JsonDeserializer<T> getDeserializer() {
      return deserializer;
    }

    public JsonStream getStream() {
      return stream;
    }
  }

  private final Map<Object, Element<?>> vault;
  private final Queue<Element<?>> queue;

  public JsonGraph() {
    this.vault = new HashMap<>();
    this.queue = new LinkedList<>();
  }

  /**
   * @param vault
   */
  public JsonGraph(Map<Object, Element<?>> vault) {
    this.vault = vault;
    this.queue = new LinkedList<>();
  }

  /**
   * @param json
   * @param instance
   * @param type
   */
  public void buildWriterGraph(Json json, T instance, Token<T> type) {
    Token<T> token = type;

    Class<?> reference = token.getRawType();
    JsonPolicy policy = json.getPolicy();
    boolean superClass = false;

    while ((reference != null) && (reference != Object.class)) {
      Collection<FieldToken> tokens = TokenFactory.stateSerializationFields(policy, token, superClass);
      for (FieldToken each : tokens) {
        Object value = each.getValue(instance);
        Token<?> valueToken = null;
        if (Objects.isNull(value)) {
          if (policy.getNullSerialization()) {
            valueToken = new Token<Object>() {
            };
          } else {
            continue;
          }
        }

        Token<?> genericSafeToken;
        if (valueToken != null) {
          genericSafeToken = valueToken;
        } else {
          Type sign = Token.resolve(token.getType(), reference, each.getField().getGenericType());
          sign = resolveWildCardType(sign);
          genericSafeToken = new Token<>(sign);
        }

        Element<?> element = getWriterElement(json, genericSafeToken, each);
        getVault().put(each.getName(), element);
        getQueue().add(element);
      }

      if (policy.getSuperClassSerialization()) {
        Type resolvedType = Token.resolve(token.getType(), reference, reference.getGenericSuperclass());
        token = new Token<>(resolvedType);
        reference = token.getRawType();
        superClass = true;
      } else {
        break;
      }
    }
  }

  /**
   * @param json
   * @param fieldType
   * @param valueType
   * @return
   */
  private Element<?> getWriterElement(final Json json, final Token<?> sign, final FieldToken fieldType) {
    Element<?> element;
    JsonSerializer<?> serializer;

    Serializer annotatedSerializer = fieldType.getSignedSerializer();
    if (annotatedSerializer != null) {
      serializer = getSignedSerializer(annotatedSerializer);
    } else {
      serializer = json.getSerializer(fieldType.getName());
    }

    if (serializer != null) {
      element = new Element<>(fieldType, null, serializer);
    } else {
      JsonAdapter<?> adapter;
      Adapter signedAdapter = fieldType.getSignedAdapter();

      if (signedAdapter != null) {
        adapter = getSignedAdapter(signedAdapter);
      } else {
        adapter = json.getAdapter(sign);
      }

      element = new Element<>(fieldType, adapter);
    }

    return element;
  }

  /**
   * @param serializer
   * @return
   */
  private JsonSerializer<?> getSignedSerializer(final Serializer serializer) {
    Class<?> serializerType = serializer.type();
    ConstructorToken constructorToken = new ConstructorToken(serializerType);
    JsonSerializer<?> instance = (JsonSerializer<?>) constructorToken.getDefaultConstructor();

    if (instance == null) {
      throw new JsonException("Annotated Serializer doesn't have default constructor."
          + "{Hint: ErrorCode: 109}");
    }

    return instance;
  }

  /**
   * @param adapter
   * @return
   */
  private JsonAdapter<?> getSignedAdapter(final Adapter adapter) {
    Class<?> adapterType = adapter.type();
    ConstructorToken token = new ConstructorToken(adapterType);
    JsonAdapter<?> adapterInstance = (JsonAdapter<?>) token.getDefaultConstructor();

    if (adapterInstance == null) {
      throw new JsonException("Annotated Adapter doesn't have default constructor."
          + "{Hint: ErrorCode: 108}");
    }

    return adapterInstance;
  }

  /**
   * @param json
   * @param stream
   * @param type
   */
  public void buildReaderGraph(Json json, JsonStream stream, Token<T> type) {
    JsonPolicy policy = json.getPolicy();

    while (stream.next()) {
      String key = stream.key();
      JsonValue value = stream.value();

      Token<?> token = type;
      Class<?> reference = token.getRawType();
      boolean superClass = false;
      FieldToken field = null;
      while (reference != null && reference != Object.class) {
        Collection<FieldToken> tokens = TokenFactory.stateDeserializationFields(policy, reference, superClass);
        boolean found = false;
        for (FieldToken next : tokens) {
          if (next.getName().equals(getKey(key, policy.getNamingPolicy()))) {
            field = next;
            found = true;
            break;
          }
        }

        if (found && field != null) {
          break;
        } else {
          if (policy.getSuperClassDeserialization()) {
            Type resolvedType = Token.resolve(token.getType(), reference, reference.getGenericSuperclass());
            token = new Token<>(resolvedType);
            reference = token.getRawType();
            superClass = true;
          } else {
            break;
          }
        }
      }

      if (field != null) {
        Element<?> element;
        if (policy.getNullDeserialization()) {
          JsonValue nullValue = new JsonNull();
          if (value.toString().equals(nullValue.toString())) {
            JsonAdapter<?> nullAdapter = json.getAdapter(new Token<Object>() {
            });
            
            element = new Element<>(field, new JsonStream(value), nullAdapter);
            getVault().put(field.getName(), element);
            getQueue().add(element);
            continue;
          }
        }

        element = getReaderDeseriazerElement(json, field, value);
        if (element == null) {
          Type sign = Token.resolve(token.getType(), reference, field.getField().getGenericType());
          sign = resolveWildCardType(sign);
          if (sign == Object.class) {
            sign = getRuntimeAdapterSign(stream);
          }

          element = getReaderAdapterElement(json, new Token<>(sign), field, value);
        }

        getVault().put(field.getName(), element);
        getQueue().add(element);
      }
    }
  }

  /**
   * @param json
   * @param fieldType
   * @param value
   * @return
   */
  private Element<?> getReaderDeseriazerElement(final Json json, final FieldToken fieldType, final JsonValue value) {
    Element<?> element = null;
    JsonDeserializer<?> deserializer;

    Deserializer annotatedDeserializer = fieldType.getSignedDeserializer();
    if (annotatedDeserializer != null) {
      deserializer = getSignedDeserializer(annotatedDeserializer);
    } else {
      deserializer = json.getDeserializer(fieldType.getName());
    }

    if (deserializer != null) {
      element = new Element<>(fieldType, new JsonStream(value), deserializer);
    }

    return element;
  }

  /**
   * @param json
   * @param sign
   * @param fieldType
   * @param value
   * @return
   */
  private Element<?> getReaderAdapterElement(final Json json, final Token<?> sign,
      final FieldToken fieldType, final JsonValue value) {
    Element<?> element;
    JsonAdapter<?> adapter;

    Adapter annotatedAdapter = fieldType.getSignedAdapter();
    if (annotatedAdapter != null) {
      adapter = getSignedAdapter(annotatedAdapter);
    } else {
      adapter = json.getAdapter(sign);
    }

    element = new Element<>(fieldType, new JsonStream(value), adapter);
    return element;
  }

  /**
   * @param deserializer
   * @return
   */
  private JsonDeserializer<?> getSignedDeserializer(final Deserializer deserializer) {
    Class<?> deserializerType = deserializer.type();
    ConstructorToken constructorToken = new ConstructorToken(deserializerType);
    JsonDeserializer<?> instance = (JsonDeserializer<?>) constructorToken.getDefaultConstructor();

    if (instance == null) {
      throw new JsonException("Annotated Deserializer doesn't have default constructor."
          + "{Hint: ErrorCode:110}");
    }

    return instance;
  }

  /**
   * @param stream
   * @return
   */
  private Type getRuntimeAdapterSign(JsonStream stream) {
    stream.peek();

    switch (stream.getStatus()) {
      case Bool:
        return Boolean.class;
      case Char:
        return Character.class;
      case Number:
        return Number.class;
      case String:
        return String.class;
    }

    return Object.class;
  }

  /**
   * @param sign
   * @return
   */
  private Type resolveWildCardType(Type sign) {
    Type resolvedType;

    if (sign instanceof WildcardType) {
      WildcardType wt = (WildcardType) sign;
      resolvedType = Token.resolveBounds(wt);

      if ((resolvedType != null) && (resolvedType instanceof Class<?>)) {
        return resolvedType;
      }
    }

    return sign;
  }

  private String getKey(String key, JsonPolicy.KeyNamingPolicy namingPolicy) {
    if (namingPolicy == null) {
      return key;
    }

    switch (namingPolicy) {
      case DEFAULT:
        return key;
      case LOWERCASE:
        return key.toLowerCase(Locale.ENGLISH);
      case UPPERCASE:
        return key.toUpperCase(Locale.ENGLISH);
    }

    return key;
  }

  /**
   * @version 1.0.3
   * @since 1.0.3
   */
  static class TokenFactory {

    /**
     * @param policy
     * @param type
     * @param superClass
     * @return
     */
    public static Collection<FieldToken> stateDeserializationFields(JsonPolicy policy, Class<?> type, boolean superClass) {
      if (policy.getIgnoreFieldModifiers()) {
        return getDeserilizationFields(policy, type);
      }

      return getDeserilizationFields(policy, type, superClass);
    }

    /**
     * @param policy
     * @param type
     * @param superClass
     * @return
     */
    private static Collection<FieldToken> getDeserilizationFields(JsonPolicy policy, Class<?> type, boolean superClass) {
      Set<Integer> modSet = getModifiers(policy, superClass);
      int[] modArray = getModifiers(modSet);

      Collection<FieldToken> allFields = Token.getDeclaredFields(type);
      Collection<FieldToken> tokens = new ArrayList<>();

      for (FieldToken token : allFields) {
        if (hasDeserializer(policy.getDeserializers(), token)) {
          tokens.add(token);
          continue;
        }

        if (token.isSignedWithAnnotation()) {
          Ignored ignoredAnnotation = token.getSignedIgnored();
          if (ignoredAnnotation == null) {
            tokens.add(token);
          } else {
            if (ignoredAnnotation.deserialize()) {
              tokens.add(token);
            }
          }

          continue;
        }

        if (isIgnored(policy, token)) {
          continue;
        }

        if (token.isEnum() && !policy.getEnumDeserialization()) {
          continue;
        }

        if (Ints.contains(modArray, token.getModifiers())) {
          tokens.add(token);
        }
      }

      return tokens;
    }

    /**
     * @param policy
     * @param type
     * @return
     */
    private static Collection<FieldToken> getDeserilizationFields(JsonPolicy policy, Class<?> type) {
      Collection<FieldToken> declaredFields = Token.getDeclaredFields(type);
      Collection<FieldToken> tokens = new LinkedList<>();

      for (FieldToken declaredField : declaredFields) {
        if (hasDeserializer(policy.getDeserializers(), declaredField)) {
          tokens.add(declaredField);
        }

        if (declaredField.isSignedWithAnnotation()) {
          Ignored ignoredAnnotation = declaredField.getSignedIgnored();
          if (ignoredAnnotation == null) {
            tokens.add(declaredField);
          } else {
            if (ignoredAnnotation.deserialize()) {
              tokens.add(declaredField);
            }
          }

          continue;
        }

        if (isIgnored(policy, declaredField)) {
          continue;
        }

        if ((declaredField.isEnum() && policy.getEnumDeserialization())) {
          tokens.add(declaredField);
        }
      }

      return tokens;
    }

    /**
     * @param policy
     * @param type
     * @param superClass
     * @return
     */
    public static Collection<FieldToken> stateSerializationFields(JsonPolicy policy, Token<?> type, boolean superClass) {
      if (policy.getIgnoreFieldModifiers()) {
        return getSerializationFields(policy, type);
      }

      return getSerializationFields(policy, type, superClass);
    }

    /**
     * <p>
     * Method filters,
     *
     * <ul>
     * <li> {@code JsonPolicy} modifier options for both base and super classes. </li>
     * <li> {@code JsonPolicy} Ignored Field names and Class Types. </li>
     * <li> {@code JsonPolicy} Enum Serialization option. </li>
     * </ul>
     * <p/>
     *
     * <p>
     * Warning: If the Field, associated with any Serializer then all filter properties defined
     * above are going to ignored. </p>
     *
     *
     * @param <T>
     * @param policy
     * @param type
     * @param superClass
     * @return
     */
    private static Collection<FieldToken> getSerializationFields(JsonPolicy policy, Token<?> type, boolean superClass) {
      Set<Integer> modSet = getModifiers(policy, superClass);
      int[] modArray = getModifiers(modSet);

      Collection<FieldToken> allFields = type.getDeclaredFields();
      Collection<FieldToken> tokens = new ArrayList<>();

      for (FieldToken token : allFields) {
        if (hasSerializer(policy.getSerializers(), token)) {
          tokens.add(token);
          continue;
        }

        if (token.isSignedWithAnnotation()) {
          Ignored ignoredAnnotation = token.getSignedIgnored();
          if (ignoredAnnotation == null) {
            tokens.add(token);
          } else {
            if (ignoredAnnotation.serialize()) {
              tokens.add(token);
            }
          }

          continue;
        }

        if (isIgnored(policy, token)) {
          continue;
        }

        if (token.isEnum() && !policy.getEnumSerialization()) {
          continue;
        }

        if (Ints.contains(modArray, token.getModifiers())) {
          tokens.add(token);
        }
      }

      return tokens;
    }

    /**
     * @param <T>
     * @param policy
     * @param type
     * @return
     */
    private static <T> Collection<FieldToken> getSerializationFields(JsonPolicy policy, Token<?> type) {
      Collection<FieldToken> declaredFields = type.getDeclaredFields();
      Collection<FieldToken> tokens = new LinkedList<>();

      // Adım Kırklareli, hiç sıkmadım düşman eli.
      for (FieldToken declaredField : declaredFields) {
        if (hasSerializer(policy.getSerializers(), declaredField)) {
          tokens.add(declaredField);
        }

        if (declaredField.isSignedWithAnnotation()) {
          Ignored ignoredAnnotation = declaredField.getSignedIgnored();
          if (ignoredAnnotation == null) {
            tokens.add(declaredField);
          } else {
            if (ignoredAnnotation.serialize()) {
              tokens.add(declaredField);
            }
          }

          continue;
        }

        if (isIgnored(policy, declaredField)) {
          continue;
        }

        if ((declaredField.isEnum() && policy.getEnumSerialization())) {
          tokens.add(declaredField);
        }
      }

      return tokens;
    }

    private static boolean hasSerializer(Map<Object, JsonSerializer<?>> map, FieldToken token) {
      return (map.get(token.getName()) != null)
          || (map.get(token.getField().getType()) != null);
    }

    private static boolean hasDeserializer(Map<Object, JsonDeserializer<?>> map, FieldToken token) {
      return (map.get(token.getName()) != null)
          || (map.get(token.getType()) != null);
    }

    private static boolean isIgnored(JsonPolicy policy, FieldToken token) {
      if (policy.getIgnoredFields().contains(token.getName())) {
        return true;
      } else if (policy.getIgnoredTypes().contains(token.getType())) {
        return true;
      }

      final Type genericType = token.getField().getGenericType();
      final Type cleanType = new Token<>(genericType).getType();
      if (policy.getIgnoredTypes().contains(cleanType)) {
        return true;
      }

      for (Type ignoredTypes : policy.getIgnoredTypes()) {
        if (ignoredTypes.equals(cleanType)) {
          return true;
        }
      }

      return false;
    }

    /**
     * @param policy
     * @param type
     * @param superClass
     * @return
     */
    private static Set<Integer> getModifiers(JsonPolicy policy, boolean superClass) {
      Set<Integer> permissions;

      if (superClass) {
        permissions = policy.getAllowedSubClassFieldModifiers();
      } else {
        permissions = policy.getAllowedFieldModifiers();
      }

      return permissions;
    }

    /**
     * @param modifiers
     * @return
     */
    private static int[] getModifiers(Set<Integer> modifiers) {
      final int[] mods;

      if (Sets.nullOrEmpty(modifiers)) {
        mods = new int[0];
      } else {
        mods = new int[modifiers.size()];
      }

      int index = 0;
      for (Integer perm : modifiers) {
        mods[index++] = perm;
      }

      return mods;
    }
  }

  /**
   * @return
   */
  public Map<Object, Element<?>> getVault() {
    return vault;
  }

  /**
   * @return
   */
  public Queue<Element<?>> getQueue() {
    return queue;
  }
}
