/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * This file is part of terrayazilim-json project.
 *
 * This file incorporates work covered by
 * the following copyright and permission notices:
 *
 * Copyright (C) 2018 Terra Yazılım Bilişim Hiz. Elek. Dan. Oto. ve Loj. Tic. Ltd. Şti. | info [at] terrayazilim [dot] com [dot] tr
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package tr.com.terrayazilim.json;

import java.io.Serializable;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.lang.reflect.Type;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import tr.com.terrayazilim.core.Objects;

/**
 * <p> {@link JsonPolicy} contains Serialization & Deserialization properties. </p>
 * 
 * <p> Properties are using lazy instantiation. See getter methods. However, default 
 * values of properties are shown below. </p>
 * <ul>
 *  <li> Default value of {@link #namingPolicy} is {@code KeyNamingPolicy.DEFAULT}. </li>
 *  <ul> Default values of {@link #allowedSubClassFieldModifiers} are:
 *    <li> {@code java.lang.Modifier.PRIVATE} </li>
 *    <li> {@code java.lang.Modifier.PROTECTED} </li>
 *    <li> {@code java.lang.Modifier.PUBLIC} </li>
 *  </ul>
 * 
 *  <ul> Default values of {@link #allowedFieldModifiers} are:
 *    <li> {@code java.lang.Modifier.PRIVATE} </li>
 *    <li> {@code java.lang.Modifier.PROTECTED} </li>
 *    <li> {@code java.lang.Modifier.PUBLIC} </li>
 *  </ul>
 * 
 *  <li> {@link #adapters} doesn't return null. It returns {@code new HashMap()} as default. </li>
 *  <li> {@link #serializers} doesn't return null. It returns {@code new HashMap()} as default. </li>
 *  <li> {@link #deserializers} doesn't return null. It returns {@code new HashMap()} as default. </li>
 *  <li> {@link #serializers} doesn't return null. It returns {@code new HashMap()} as default. </li>
 *  <li> {@link #ignoredFields} doesn't return null. It returns {@code new HashSet()} as default. </li>
 *  <li> {@link #ignoredTypes} doesn't return null. It returns {@code new HashSet()} as default. </li>
 *  <li> {@link #ignoreFieldModifiers} set {@code true} by default. </li>
 *  <li> Default value of {@link #superClassSerialization} is {@code true} </li>
 *  <li> Default value of {@link #superClassDeserialization} is {@code true} </li>
 *  <li> Default value of {@link #enumSerialization} is {@code true} </li>
 *  <li> Default value of {@link #enumDeserialization} is {@code false} </li>
 *  <li> Default value of {@link #nullSerialization} is {@code false} </li>
 *  <li> Default value of {@link #dateFormat} is {@literal dd-MM-yyy HH:mm:ss} </li>
 * </ul>
 * 
 * <p>
 *  Usage Notes:
 * </p>
 * 
 * <p> Both {@link #deserializers} and {@link #serializers} keys represents 
 * either field name or class type. Fields that associated with that key are won't be
 * filtered by {@link #ignoredFields}, {@link #ignoredTypes}, {@link #allowedFieldModifiers},
 * {@link #allowedSubClassFieldModifiers}. Since {@link #deserializers} and {@link #serializers}
 * keys doesn't get affected from above settings, it's still viable to 
 * {@link #superClassSerialization} and {@link #superClassDeserialization}.
 * </p>
 * 
 * <p>
 *  If {@link #ignoreFieldModifiers} is set {@code true}, then {@link #allowedFieldModifiers} and
 * {@link #allowedSubClassFieldModifiers} become redundant. If the class that you want to serialize or
 * deserialize is simple or doesn't need any {@link Field} {@link java.lang.reflect.Modifier} filter, 
 * then use this property to gain minor performance impact.
 * </p>
 * 
 * @author M.Çağrı Tepebaşılı - cagritepebasili [at] protonmail [dot] com
 * @version 1.0.3
 * @since 1.0.3
 */
public class JsonPolicy implements Serializable {

  /**
   * @version 1.0.3
   * @since 1.0.3
   */
  static public enum KeyNamingPolicy {

    DEFAULT,
    LOWERCASE,
    UPPERCASE;
  }

  private KeyNamingPolicy namingPolicy = KeyNamingPolicy.DEFAULT;

  private Set<Integer> allowedSubClassFieldModifiers;
  private Set<Integer> allowedFieldModifiers;

  private Map<Type, JsonAdapter<?>> adapters;
  
  /**
   * {@code Map.key} is either fieldName or Type.
   */
  private Map<Object, JsonSerializer<?>> serializers;
  
  /**
   * {@code Map.key} is either fieldName or Type.
   */
  private Map<Object, JsonDeserializer<?>> deserializers;
  
  private Set<String> ignoredFields;
  private Set<Type> ignoredTypes;
  private Boolean ignoreFieldModifiers;

  private Boolean superClassSerialization;
  private Boolean superClassDeserialization;
  private Boolean enumSerialization;
  private Boolean enumDeserialization;
  
  private Boolean nullSerialization;
  private Boolean nullDeserialization;
  
  private String dateFormat;

  public JsonPolicy() {
  }
  
  public JsonPolicy(JsonPolicy policy) {
    this.namingPolicy = policy.getNamingPolicy();
    this.allowedFieldModifiers = policy.getAllowedFieldModifiers();
    this.allowedSubClassFieldModifiers = policy.getAllowedSubClassFieldModifiers();
    this.adapters = policy.getAdapters();
    this.serializers = policy.getSerializers();
    this.deserializers = policy.getDeserializers();
    this.ignoredFields = policy.getIgnoredFields();
    this.ignoredTypes = policy.getIgnoredTypes();
    this.superClassSerialization = policy.getSuperClassSerialization();
    this.superClassDeserialization = policy.getSuperClassDeserialization();
    this.enumSerialization = policy.getEnumSerialization();
    this.enumDeserialization = policy.getEnumDeserialization();
    this.nullSerialization = policy.getNullSerialization();
    this.nullDeserialization = policy.getNullDeserialization();
    this.dateFormat = policy.getDateFormat();
  }

  /**
   * @return
   */
  public KeyNamingPolicy getNamingPolicy() {
    if (Objects.isNull(this.namingPolicy)) {
      this.namingPolicy = KeyNamingPolicy.DEFAULT;
    }

    return namingPolicy;
  }

  /**
   * @param namingPolicy
   */
  public void setNamingPolicy(KeyNamingPolicy namingPolicy) {
    this.namingPolicy = namingPolicy;
  }

  /**
   * @return
   */
  public Set<Integer> getAllowedSubClassFieldModifiers() {
    if (Objects.isNull(this.allowedSubClassFieldModifiers)) {
      this.allowedSubClassFieldModifiers = new HashSet<>();
      
      final int[] mod = new int[]{
        Modifier.PRIVATE, Modifier.PROTECTED, Modifier.PUBLIC
      };

      for (int modifier : mod) {
        allowedSubClassFieldModifiers.add(modifier);
      }

      allowedSubClassFieldModifiers.add(17);
    }

    return allowedSubClassFieldModifiers;
  }

  /**
   * @param allowedSubClassFieldModifiers
   */
  public void setAllowedSubClassFieldModifiers(Set<Integer> allowedSubClassFieldModifiers) {
    this.allowedSubClassFieldModifiers = allowedSubClassFieldModifiers;
  }

  /**
   * @return
   */
  public Set<Integer> getAllowedFieldModifiers() {
    if (Objects.isNull(this.allowedFieldModifiers)) {
      this.allowedFieldModifiers = new HashSet<>();
      
      final int[] mod = new int[]{
        Modifier.PRIVATE, Modifier.PROTECTED, Modifier.PUBLIC
      };

      for (int modifier : mod) {
        allowedFieldModifiers.add(modifier);
      }

      allowedFieldModifiers.add(17);
    }
    
    return allowedFieldModifiers;
  }

  /**
   * @param allowedFieldModifiers
   */
  public void setAllowedFieldModifiers(Set<Integer> allowedFieldModifiers) {
    this.allowedFieldModifiers = allowedFieldModifiers;
  }

  /**
   * @return
   */
  public Map<Type, JsonAdapter<?>> getAdapters() {
    if (Objects.isNull(this.adapters)) {
      this.adapters = new HashMap<>();
    }
    
    return adapters;
  }

  /**
   * @param adapters
   */
  public void setAdapters(Map<Type, JsonAdapter<?>> adapters) {
    this.adapters = adapters;
  }

  /**
   * @return
   */
  public Map<Object, JsonSerializer<?>> getSerializers() {
    if (Objects.isNull(this.serializers)) {
      this.serializers = new HashMap<>();
    }
    
    return serializers;
  }

  /**
   * @param codecs
   */
  public void setSerializer(Map<Object, JsonSerializer<?>> codecs) {
    this.serializers = codecs;
  }

  /**
   * @return 
   */
  public Map<Object, JsonDeserializer<?>> getDeserializers() {
    if (Objects.isNull(this.deserializers)) {
      this.deserializers = new HashMap<>();
    }
    
    return deserializers;
  }

  /**
   * @param deserializers 
   */
  public void setDeserializers(Map<Object, JsonDeserializer<?>> deserializers) {
    this.deserializers = deserializers;
  }

  /**
   * @return
   */
  public Set<String> getIgnoredFields() {
    if (Objects.isNull(this.ignoredFields)) {
      this.ignoredFields = new HashSet<>();
    }
    
    return ignoredFields;
  }

  /**
   * @param ignoredFields
   */
  public void setIgnoredFields(Set<String> ignoredFields) {
    this.ignoredFields = ignoredFields;
  }

  /**
   * @return 
   */
  public Set<Type> getIgnoredTypes() {
    if (Objects.isNull(this.ignoredTypes)) {
      this.ignoredTypes = new HashSet<>();
    }
    
    return ignoredTypes;
  }

  /**
   * @param ignoredTypes 
   */
  public void setIgnoredTypes(Set<Type> ignoredTypes) {
    this.ignoredTypes = ignoredTypes;
  }

  /**
   * @return 
   */
  public Boolean getIgnoreFieldModifiers() {
    if (this.ignoreFieldModifiers == null) {
      this.ignoreFieldModifiers = Boolean.FALSE;
    }
    
    return ignoreFieldModifiers;
  }

  /**
   * @param ignoreFieldModifiers 
   */
  public void setIgnoreFieldModifiers(Boolean ignoreFieldModifiers) {
    this.ignoreFieldModifiers = ignoreFieldModifiers;
  }

  /**
   * @return
   */
  public Boolean getSuperClassSerialization() {
    if (Objects.isNull(this.superClassSerialization)) {
      this.superClassSerialization = Boolean.TRUE;
    }
    
    return superClassSerialization;
  }

  /**
   * @param superClassSerialization
   */
  public void setSuperClassSerialization(Boolean superClassSerialization) {
    this.superClassSerialization = superClassSerialization;
  }

  /**
   * @return
   */
  public Boolean getSuperClassDeserialization() {
    if (Objects.isNull(this.superClassDeserialization)) {
      this.superClassDeserialization = Boolean.TRUE;
    }
    
    return superClassDeserialization;
  }

  /**
   * @param superClassDeserialization
   */
  public void setSuperClassDeserialization(Boolean superClassDeserialization) {
    this.superClassDeserialization = superClassDeserialization;
  }

  /**
   * @return
   */
  public Boolean getEnumSerialization() {
    if (Objects.isNull(this.enumSerialization)) {
      this.enumSerialization = Boolean.TRUE;
    }
    
    return enumSerialization;
  }

  /**
   * @param enumSerialization
   */
  public void setEnumSerialization(Boolean enumSerialization) {
    this.enumSerialization = enumSerialization;
  }

  /**
   * @return
   */
  public Boolean getEnumDeserialization() {
    if (Objects.isNull(this.enumDeserialization)) {
      this.enumDeserialization = Boolean.FALSE;
    }
    
    return enumDeserialization;
  }

  /**
   * @param enumDeserialization
   */
  public void setEnumDeserialization(Boolean enumDeserialization) {
    this.enumDeserialization = enumDeserialization;
  }

  /**
   * @return
   */
  public Boolean getNullSerialization() {
    if (Objects.isNull(this.nullSerialization)) {
      this.nullSerialization = Boolean.FALSE;
    }
    
    return nullSerialization;
  }

  /**
   * @param nullSerialization
   */
  public void setNullSerialization(Boolean nullSerialization) {
    this.nullSerialization = nullSerialization;
  }

  public Boolean getNullDeserialization() {
    if (this.nullDeserialization == null) {
      this.nullDeserialization = Boolean.TRUE;
    }
    
    return nullDeserialization;
  }

  public void setNullDeserialization(Boolean nullDeserialization) {
    this.nullDeserialization = nullDeserialization;
  }

  /**
   * @return
   */
  public String getDateFormat() {
    if (Objects.isNull(this.dateFormat)) {
      this.dateFormat = "dd-MM-yyy HH:mm:ss";
    }
    
    return dateFormat;
  }

  /**
   * @param dateFormat
   */
  public void setDateFormat(String dateFormat) {
    this.dateFormat = dateFormat;
  }
}
