/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * This file is part of terrayazilim-json project.
 *
 * This file incorporates work covered by
 * the following copyright and permission notices:
 *
 * Copyright (C) 2018 Terra Yazılım Bilişim Hiz. Elek. Dan. Oto. ve Loj. Tic. Ltd. Şti. | info [at] terrayazilim [dot] com [dot] tr
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package tr.com.terrayazilim.json.codec;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.Map;
import java.util.Properties;
import tr.com.terrayazilim.json.Json;
import tr.com.terrayazilim.json.JsonAdapter;
import tr.com.terrayazilim.json.JsonAdapterFactory;
import tr.com.terrayazilim.json.JsonException;
import tr.com.terrayazilim.json.stream.JsonBuffer;
import tr.com.terrayazilim.json.stream.JsonStream;
import tr.com.terrayazilim.json.type.ConstructorToken;
import tr.com.terrayazilim.json.type.Token;

/**
 * @author M.Çağrı Tepebaşılı - cagritepebasili [at] protonmail [dot] com
 * @version 1.0.3
 * @since 1.0.3
 * @param <K>
 * @param <V>
 */
public class JsonMapAdapter<K, V> extends JsonAdapter<Map<K, V>> {

  private final JsonAdapterFactory adapterFactory;
  private JsonAdapter<K> keyAdapter;
  private JsonAdapter<V> valueAdapter;

  private final Token<Map<K, V>> token;
  private final Json json;

  /**
   * @param json
   * @param tokenArgs
   */
  public JsonMapAdapter(Json json, final Token<Map<K, V>> tokenArgs) {
    this.json = json;
    this.token = tokenArgs;

    this.adapterFactory = new JsonAdapterFactory() {
      @Override
      public <T> JsonAdapter<T> call(Json json, final Token<T> token) {
        Class<?> sourceRaw = Token.rawTypeOf(token.getType());
        Type[] types = getTypes(token.getType(), sourceRaw);

        Type resolvedKeyType = Token.resolveWildCardType(types[0]);
        Type resolvedValueType = Token.resolveWildCardType(types[1]);

        keyAdapter = (JsonAdapter<K>) json.getAdapter(new Token<>(resolvedKeyType));
        valueAdapter = (JsonAdapter<V>) json.getAdapter(new Token<>(resolvedValueType));

        return (JsonAdapter<T>) new JsonAdapter<Map<K, V>>() {
          @Override
          public void write(JsonBuffer buffer, Map<K, V> object) {
            buffer.beginArray();

            for (Map.Entry<K, V> entry : object.entrySet()) {
              buffer.beginArray();
              keyAdapter.write(buffer, entry.getKey());
              valueAdapter.write(buffer, entry.getValue());
              buffer.endArray();
            }

            buffer.endArray();
          }

          @Override
          public Map<K, V> read(JsonStream stream) {
            Map<K, V> map = (Map<K, V>) ConstructorToken.getMap(token);
            
            requireArray(stream.getStatus());
            while (stream.next()) {
              JsonStream pair = new JsonStream(stream.value());
              K key = keyAdapter.read(new JsonStream(pair.nextValue()));
              V value = valueAdapter.read(new JsonStream(pair.nextValue()));
              
              map.put(key, value);
            }

            return map;
          }
        };
      }
    };
  }

  @Override
  public void write(JsonBuffer buffer, Map<K, V> object) {
    adapterFactory.call(json, token).write(buffer, object);
  }

  @Override
  public Map<K, V> read(JsonStream stream) {
    return adapterFactory.call(json, token).read(stream);
  }

  static void requireArray(JsonStream.Status status) {
    if (!status.equals(JsonStream.Status.Array)) {
      throw new JsonException("Json Format is not acceptable for java.util.Map."
          + "{Hint: ErrorCode:121}");
    }
  }

  static Type[] getTypes(Type context, Class<?> contextRawType) {
    if (context == Properties.class) {
      return new Type[]{String.class, String.class};
    }

    Type mapType = Token.getSupertype(context, contextRawType, Map.class);
    if (mapType instanceof ParameterizedType) {
      ParameterizedType mapParameterizedType = (ParameterizedType) mapType;
      return mapParameterizedType.getActualTypeArguments();
    }
    return new Type[]{Object.class, Object.class};
  }
}
