/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * This file is part of terrayazilim-jsonObject project.
 *
 * This file incorporates work covered by
 * the following copyright and permission notices:
 *
 * Copyright (C) 2018 Terra Yazılım Bilişim Hiz. Elek. Dan. Oto. ve Loj. Tic. Ltd. Şti. | info [at] terrayazilim [dot] com [dot] tr
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package tr.com.terrayazilim.json.stream;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import tr.com.terrayazilim.core.Objects;
import tr.com.terrayazilim.core.tuple.NativeEntry;
import tr.com.terrayazilim.json.JsonArray;
import tr.com.terrayazilim.json.JsonException;
import tr.com.terrayazilim.json.JsonLiteral;
import tr.com.terrayazilim.json.JsonNull;
import tr.com.terrayazilim.json.JsonObject;
import tr.com.terrayazilim.json.JsonValue;

/**
 * @author M.Çağrı Tepebaşılı - cagritepebasili [at] protonmail [dot] com
 * @version 1.0.3
 * @since 1.0.3
 */
public final class JsonStream {

  static public enum Status {

    Begin,
    JsonNull,
    Object,
    Array,
    Number,
    Bool,
    Char,
    String,
    End;
    
  }
  
  private Queue<Map.Entry<String, JsonValue>> queue;
  private Map.Entry<String, JsonValue> entry;
  
  private final JsonObject jsonObject;
  private final JsonArray jsonArray;
  private final JsonLiteral jsonLiteral;

  private Status status;

  /**
   * @param jsonValue 
   */
  public JsonStream(JsonValue jsonValue) {
    if (jsonValue instanceof JsonLiteral) {
      this.jsonLiteral = (JsonLiteral) jsonValue;
      this.jsonArray = null;
      this.jsonObject = null;
    } else if (jsonValue instanceof JsonArray) {
      this.jsonArray = (JsonArray) jsonValue;
      this.jsonLiteral = null;
      this.jsonObject = null;
    } else if (jsonValue instanceof JsonObject) {
      this.jsonObject = (JsonObject) jsonValue;
      this.jsonLiteral = null;
      this.jsonArray = null;
    } else if (jsonValue instanceof JsonNull) {
      Object tmp = null;
      this.jsonLiteral = new JsonLiteral(tmp);
      this.jsonArray = null;
      this.jsonObject = null;
    } else {
      throw new JsonException("Unrecognized JsonValue. {Hint: ErrorCode: 113}");
    }
    
    resetStatus(jsonValue);
  }

  /**
   * @param jsonLiteral
   */
  public JsonStream(JsonLiteral jsonLiteral) {
    this.jsonObject = null;
    this.jsonArray = null;
    this.jsonLiteral = jsonLiteral;
    resetStatus(jsonLiteral);
  }

  /**
   * @param json
   */
  public JsonStream(JsonObject json) {
    this.jsonArray = null;
    this.jsonLiteral = null;
    this.jsonObject = json;
    this.status = Status.Object;
  }

  /**
   * @param jsonArray
   */
  public JsonStream(JsonArray jsonArray) {
    this.jsonObject = null;
    this.jsonLiteral = null;
    this.jsonArray = jsonArray;
    this.status = Status.Array;
  }

  /**
   * @return
   */
  private Queue<Map.Entry<String, JsonValue>> getQueue() {
    if (this.queue == null) {
      this.queue = new LinkedList<>();
      String nullString = null;
      if (isObjectStream()) {
        final Iterator<String> it = jsonObject.keys();
        while (it.hasNext()) {
          String next = it.next();
          Object value = jsonObject.get(next);
          JsonValue jval = null;
          if (JsonLiteral.isLiteral(value)) {
            jval = new JsonLiteral(value);
          } else if (value instanceof JsonObject) {
            jval = (JsonObject) value;
          } else if (value instanceof JsonArray) {
            jval = (JsonArray) value;
          } else {
            throw new JsonException("Found object is not instance of JsonLiteral, JsonObject or JsonArray."
                + "{Hint: ErrorCode:114}");
          }

          queue.add(NativeEntry.of(next, jval));
        }
      } else if (isArrayStream()) {
        final Iterator<Object> it = jsonArray.iterator();
        while (it.hasNext()) {
          Object next = it.next();
          JsonValue jval = null;
          if (JsonLiteral.isLiteral(next)) {
            jval = new JsonLiteral(next);
          } else if (next instanceof JsonObject) {
            jval = (JsonObject) next;
          } else if (next instanceof JsonArray) {
            jval = (JsonArray) next;
          }

          queue.add(NativeEntry.of(nullString, jval));
        }
      } else if (isLiteralStream()) {
        JsonValue jval = jsonLiteral;
        queue.add(NativeEntry.of(nullString, jval));
      }
    }

    return queue;
  }
  
  /**
   * Method determines the Status of given {@code JsonLiteral}.
   * 
   * @param literal 
   */
  private void resetStatus(JsonLiteral literal) {
     if (literal.getValue() == null) {
       setStatus(Status.JsonNull);
    } else if (literal.getValue() instanceof Number) {
      setStatus(Status.Number);
    } else if (literal.getValue() instanceof Boolean) {
      setStatus(Status.Bool);
    } else if (literal.getValue() instanceof Character) {
      setStatus(Status.Char);
    } else if (literal.getValue() instanceof String) {
      setStatus(Status.String);
    }
  }
  
  /**
   * Method determines the Status of given {@code JsonValue}.
   * 
   * @param value 
   */
  private void resetStatus(JsonValue value) {
    if (value instanceof JsonNull) {
      setStatus(Status.JsonNull);
    } else if (value instanceof JsonLiteral) {
      resetStatus(((JsonLiteral) value));
    } else if (value instanceof JsonObject) {
      setStatus(Status.Object);
    } else if (value instanceof JsonArray) {
      setStatus(Status.Array);
    } else {
      throw new JsonException("Unspotted Status. {Hint: ErrorCode:115}");
    }
  }

  /**
   * @return
   */
  public boolean hasNext() {
    return !getQueue().isEmpty();
  }

  /**
   * @return Key. Null if entry haven't been set or key is null.
   */
  public String key() {
    if (Objects.isNull(getEntry())) {
      return null;
    }

    return getEntry().getKey();
  }

  /**
   * @return Value. Null if entry haven't been set or value is null.
   */
  public JsonValue value() {
    if (Objects.isNull(getEntry())) {
      return null;
    }

    return getEntry().getValue();
  }

  /**
   * Method peeks the Queue and set {@link #getEntry()}. After this method called,
   * {@link #key()}, {@link #value()} and other method connected to {@link #entry} can be
   * used.
   */
  public void begin() {
    Map.Entry<String, JsonValue> next = getQueue().peek();
    setEntry(next);
    
    // Override Status to Status.Begin
    setStatus(Status.Begin);
  }
  
  public JsonValue peek() {
    Map.Entry<String, JsonValue> next = getQueue().peek();
    
    // set current Entry and spot status.
    setEntry(next);
    
    return next.getValue();
  }
  
  /**
   * Method polls the Queue and set {@link #getEntry()}. After this method called,
   * {@link #key()}, {@link #value()} can be used.
   *
   * @return true if poll succeed. Otherwise false.
   */
  public boolean next() {
    Map.Entry<String, JsonValue> next = getQueue().poll();
    if (Objects.nonNull(next)) {
      
      // set polled entry and spot status.
      setEntry(next);
      return true;
    }

    return false;
  }

  public String nextString() {
    if (!next()) {
      throw new JsonException("There is no more element in JsonStream. {Hint: ErrorCode:116}");
    }

    return value().toString();
  }
  
  public String peekString() {
    peek();
    
    return value().toString();
  }

  /**
   * If next element is not type of int, method throws {@code NumberFormatException}.
   *
   * @return next element as int.
   */
  public int nextInt() {
    return new BigDecimal(nextString()).intValue();
  }
  
  public int peekInt() {
    return new BigDecimal(peekString()).intValue();
  }

  /**
   * If next element is not type of double, method throws {@code NumberFormatException}.
   *
   * @return next element as double.
   */
  public double nextDouble() {
    return new BigDecimal(nextString()).doubleValue();
  }
  
  public double peekDouble() {
    return new BigDecimal(peekString()).doubleValue();
  }

  /**
   * If next element is not type of long, method throws {@code NumberFormatException}.
   *
   * @return next element as long.
   */
  public long nextLong() {
    return new BigDecimal(nextString()).longValue();
  }
  
  public long peekLong() {
    return new BigDecimal(peekString()).longValue();
  }

  /**
   * If next element is not type of float, method throws {@code NumberFormatException}.
   *
   * @return next element as float.
   */
  public float nextFloat() {
    return new BigDecimal(nextString()).floatValue();
  }
  
  public float peekFloat() {
    return new BigDecimal(peekString()).floatValue();
  }

  /**
   * If next element is not type of short, method throws {@code NumberFormatException}.
   *
   * @return next element as short.
   */
  public short nextShort() {
    return new BigDecimal(nextString()).shortValue();
  }
  
  public short peekShort() {
    return new BigDecimal(peekString()).shortValue();
  }

  /**
   * If next element is not type of byte, method throws {@code NumberFormatException}.
   *
   * @return next element as byte.
   */
  public byte nextByte() {
    return new BigDecimal(nextString()).byteValue();
  }
  
  public byte peekByte() {
    return new BigDecimal(peekString()).byteValue();
  }

  /**
   * If there is no more element or next element type is not instance of char, method throws
   * {@code JsonException}
   *
   * @return Next element as char.
   */
  public char nextChar() {
    if (!next()) {
      throw new JsonException("There is no more element in JsonStream. {Hint: ErrorCode:117}");
    }

    if (value() instanceof JsonLiteral) {
      JsonLiteral literal = (JsonLiteral) value();
      if (literal.getValue() instanceof Character) {
        return ((Character) literal.getValue());
      }
    }

    throw new JsonException("Next value of JsonStream is not instance of Character. {Hint: ErrorCode:118}");
  }

  /**
   * If there is no more element or next element type is not instance of Boolean, method throws
   * {@code JsonException}
   *
   * @return Next element as {@code boolean}.
   */
  public boolean nextBool() {
    if (!next()) {
      throw new JsonException("There is no more element in JsonStream. {Hint: ErrorCode:117}");
    }

    if (value() instanceof JsonLiteral) {
      JsonLiteral literal = (JsonLiteral) value();
      if (literal.getValue() instanceof Boolean) {
        return ((Boolean) literal.getValue());
      }
    }

    throw new JsonException("Next value of JsonStream is not instance of Boolean. {Hint: ErrorCode:119}");
  }
  
  /**
   * If there is no more element or next element type is not instance of JsonArray, method throws
   * {@code JsonException}
   *
   * @return Next element as {@code JsonArray}.
   */
  public JsonArray nextArray() {
    if (!next()) {
      throw new JsonException("There is no more element in JsonStream. {Hint: ErrorCode:117}");
    }
    
    if (value() instanceof JsonArray) {
      JsonArray array = (JsonArray) value();
      return array;
    }
    
    throw new JsonException("Next value of JsonStream is not instance of JsonArray. {Hint: ErrorCode:120}");
  }
  
  /**
   * If peeked element type is not instance of JsonArray, method throws
   * {@code JsonException}
   *
   * @return Peeked element as {@code JsonArray}.
   */
  public JsonArray peekArray() {
    if (value() instanceof JsonArray) {
      JsonArray array = (JsonArray) value();
      return array;
    }
    
    throw new JsonException("Peek value of JsonStream is not instance of JsonArray. {Hint: ErrorCode:121}");
  }

  public String nextKey() {
    if (!next()) {
      throw new JsonException("There is no more element in JsonStream. {Hint: ErrorCode:117}");
    }
    
    return key();
  }
  
  public JsonValue nextValue() {
    if (!next()) {
      throw new JsonException("There is no more element in JsonStream. {Hint: ErrorCode:117}");
    }
    
    return value();
  }
  
  /**
   * Ends the stream. Method sets null providers. After called this method;
   * {@link #key()}, {@link #value()} etc shouldn't be used.
   */
  public void end() {
    setEntry(null);
    setQueue(null);
    
    // Override status to Status.End.
    setStatus(Status.End);
  }

  /**
   * @return
   */
  public Iterator<JsonValue> values() {
    if (isLiteralStream()) {
      List<JsonValue> stack = new ArrayList<>(1);
      stack.add(getJsonLiteral());
      return stack.iterator();
    }

    final LinkedList<JsonValue> stack = new LinkedList<>();
    final Iterator<Map.Entry<String, JsonValue>> it = getQueue().iterator();

    while (it.hasNext()) {
      Map.Entry<String, JsonValue> next = it.next();
      stack.add(next.getValue());
    }

    return stack.iterator();
  }

  /**
   * @return
   */
  public Iterator<String> keys() {
    if (isLiteralStream()) {
      return Collections.emptyIterator();
    } else if (isArrayStream()) {
      return Collections.emptyIterator();
    }

    final LinkedList<String> stack = new LinkedList<>();
    final Iterator<Map.Entry<String, JsonValue>> it = getQueue().iterator();

    while (it.hasNext()) {
      Map.Entry<String, JsonValue> next = it.next();
      stack.add(next.getKey());
    }

    return stack.iterator();
  }
  
  /**
   * @return 
   */
  public int size() {
    return getQueue().size();
  }

  private void setEntry(Map.Entry<String, JsonValue> entry) {
    this.entry = entry;
    
    // handle status.
    if (this.entry != null) {
      resetStatus(getEntry().getValue());
    }
  }

  private Map.Entry<String, JsonValue> getEntry() {
    return this.entry;
  }

  private void setQueue(Queue<Map.Entry<String, JsonValue>> queue) {
    this.queue = queue;
  }

  private void setStatus(Status status) {
    this.status = status;
  }

  /**
   * @return Identifier for the next element type.
   */
  public Status getStatus() {
    return status;
  }

  /**
   * @return
   */
  public JsonObject getJsonObject() {
    return this.jsonObject;
  }

  /**
   * @return
   */
  public JsonArray getJsonArray() {
    return this.jsonArray;
  }

  /**
   * @return
   */
  public JsonLiteral getJsonLiteral() {
    return this.jsonLiteral;
  }

  /**
   * @return
   */
  public boolean isObjectStream() {
    return Objects.nonNull(getJsonObject());
  }

  /**
   * @return
   */
  public boolean isArrayStream() {
    return Objects.nonNull(getJsonArray());
  }

  /**
   * @return
   */
  public boolean isLiteralStream() {
    return Objects.nonNull(getJsonLiteral());
  }
}
