/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * This file is part of terrayazilim-json project.
 *
 * This file incorporates work covered by
 * the following copyright and permission notices:
 *
 * Copyright (C) 2018 Terra Yazılım Bilişim Hiz. Elek. Dan. Oto. ve Loj. Tic. Ltd. Şti. | info [at] terrayazilim [dot] com [dot] tr
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package tr.com.terrayazilim.json.stream;

/*
 Copyright (c) 2002 JSON.org

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.

 The Software shall be used for Good, not Evil.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
 */
import java.io.IOException;
import java.math.BigDecimal;
import java.util.Collection;
import java.util.Map;
import tr.com.terrayazilim.json.JsonArray;
import tr.com.terrayazilim.json.JsonException;
import tr.com.terrayazilim.json.JsonObject;

/**
 * JsonWriter provides a quick and convenient way of producing JSON text. The texts produced
 * strictly conform to JSON syntax rules. No whitespace is added, so the results are ready for
 * transmission or storage. Each instance of JsonWriter can produce one JSON text.
 * <p>
 * A JsonWriter instance provides a <code>value</code> method for appending values to the text, and
 * a <code>key</code> method for adding keys before values in objects. There are
 * <code>beginArray</code> and <code>endArray</code> methods that make and bound beginArray values,
 * and <code>beginObject</code> and <code>endObject</code> methods which make and bound beginObject
 * values. All of these methods return the JsonWriter instance, permitting a cascade style. For
 * example,
 * <pre>
 * new JsonWriter(myWriter)
 * .beginObject()
 * .key("JSON")
 * .value("Hello, World!")
 * .endObject();</pre> which writes
 * <pre>
 * {"JSON":"Hello, World!"}</pre>
 * <p>
 * The first method called must be <code>beginArray</code> or <code>beginObject</code>. There are no
 * methods for adding commas or colons. JsonWriter adds them for you. Objects and arrays can be
 * nested up to 200 levels deep.
 * <p>
 * This can sometimes be easier than using a JsonObject to build a string.
 *
 * @author JSON.org
 * @version 2016-08-08
 *
 * @author @author M.Çağrı Tepebaşılı - cagritepebasili [at] protonmail [dot] com
 * @version 1.0.3
 * @since 1.0.0
 */
public class JsonWriter {

  private static final int maxdepth = 200;

  /**
   * The comma flag determines if a comma should be output before the next value.
   */
  private boolean comma;

  /**
   * The current mode. Values: 'a' (beginArray), 'd' (done), 'i' (initial), 'k' (key), 'o'
   * (beginObject).
   */
  protected char mode;

  /**
   * The beginObject/beginArray stack.
   */
  private final JsonObject stack[];

  /**
   * The stack top index. A value of 0 indicates that the stack is empty.
   */
  private int top;

  /**
   * The writer that will receive the output.
   */
  protected Appendable writer;

  /**
   * Make a fresh JSONWriter. It can be used to build one JSON text.
   *
   * @param w
   */
  public JsonWriter(Appendable w) {
    this.comma = false;
    this.mode = 'i';
    this.stack = new JsonObject[maxdepth];
    this.top = 0;
    this.writer = w;
  }

  /**
   * Append a value.
   *
   * @param string A string value.
   * @return this
   * @throws JsonException If the value is out of sequence.
   */
  private JsonWriter append(String string) throws JsonException {
    if (string == null) {
      throw new JsonException("Null pointer");
    }
    if (this.mode == 'o' || this.mode == 'a') {
      try {
        if (this.comma && this.mode == 'a') {
          this.writer.append(',');
        }
        this.writer.append(string);
      } catch (IOException e) {
        // Android as of API 25 does not support this exception constructor
        // however we won't worry about it. If an exception is happening here
        // it will just throw a "Method not found" exception instead.
        throw new JsonException(e);
      }
      if (this.mode == 'o') {
        this.mode = 'k';
      }
      this.comma = true;
      return this;
    }
    throw new JsonException("Value out of sequence.");
  }

  /**
   * Begin appending a new beginArray. All values until the balancing <code>endArray</code> will be
   * appended to this beginArray. The <code>endArray</code> method must be called to mark the
   * beginArray's end.
   *
   * @return this
   * @throws JsonException If the nesting is too deep, or if the beginObject is started in the wrong
   * place (for example as a key or after the end of the outermost beginArray or beginObject).
   */
  public JsonWriter beginArray() throws JsonException {
    if (this.mode == 'i' || this.mode == 'o' || this.mode == 'a') {
      this.push(null);
      this.append("[");
      this.comma = false;
      return this;
    }
    throw new JsonException("Misplaced array.");
  }

  /**
   * End something.
   *
   * @param m Mode
   * @param c Closing character
   * @return this
   * @throws JsonException If unbalanced.
   */
  private JsonWriter end(char m, char c) throws JsonException {
    if (this.mode != m) {
      throw new JsonException(m == 'a'
          ? "Misplaced endArray."
          : "Misplaced endObject.");
    }
    this.pop(m);
    try {
      this.writer.append(c);
    } catch (IOException e) {
      // Android as of API 25 does not support this exception constructor
      // however we won't worry about it. If an exception is happening here
      // it will just throw a "Method not found" exception instead.
      throw new JsonException(e);
    }
    this.comma = true;
    return this;
  }

  /**
   * End an beginArray. This method most be called to balance calls to <code>beginArray</code>.
   *
   * @return this
   * @throws JsonException If incorrectly nested.
   */
  public JsonWriter endArray() throws JsonException {
    return this.end('a', ']');
  }

  /**
   * End an beginObject. This method most be called to balance calls to <code>beginObject</code>.
   *
   * @return this
   * @throws JsonException If incorrectly nested.
   */
  public JsonWriter endObject() throws JsonException {
    return this.end('k', '}');
  }

  /**
   * Append a key. The key will be associated with the next value. In an beginObject, every value
   * must be preceded by a key.
   *
   * @param string A key string.
   * @return this
   * @throws JsonException If the key is out of place. For example, keys do not belong in arrays or
   * if the key is null.
   */
  public JsonWriter key(String string) throws JsonException {
    if (string == null) {
      throw new JsonException("Null key.");
    }
    if (this.mode == 'k') {
      try {
        JsonObject topObject = this.stack[this.top - 1];
        // don't use the built in putOnce method to maintain Android support
        if (topObject.has(string)) {
          throw new JsonException("Duplicate key \"" + string + "\"");
        }
        topObject.put(string, true);
        if (this.comma) {
          this.writer.append(',');
        }
        this.writer.append(JsonObject.quote(string));
        this.writer.append(':');
        this.comma = false;
        this.mode = 'o';
        return this;
      } catch (IOException e) {
        // Android as of API 25 does not support this exception constructor
        // however we won't worry about it. If an exception is happening here
        // it will just throw a "Method not found" exception instead.
        throw new JsonException(e);
      }
    }
    throw new JsonException("Misplaced key.");
  }

  /**
   * Begin appending a new beginObject. All keys and values until the balancing
   * <code>endObject</code> will be appended to this beginObject. The <code>endObject</code> method
   * must be called to mark the beginObject's end.
   *
   * @return this
   * @throws JsonException If the nesting is too deep, or if the beginObject is started in the wrong
   * place (for example as a key or after the end of the outermost beginArray or beginObject).
   */
  public JsonWriter beginObject() throws JsonException {
    if (this.mode == 'i') {
      this.mode = 'o';
    }
    if (this.mode == 'o' || this.mode == 'a') {
      this.append("{");
      this.push(new JsonObject());
      this.comma = false;
      return this;
    }
    throw new JsonException("Misplaced object.");

  }

  /**
   * Pop an beginArray or beginObject scope.
   *
   * @param c The scope to close.
   * @throws JsonException If nesting is wrong.
   */
  private void pop(char c) throws JsonException {
    if (this.top <= 0) {
      throw new JsonException("Nesting error.");
    }
    char m = this.stack[this.top - 1] == null ? 'a' : 'k';
    if (m != c) {
      throw new JsonException("Nesting error.");
    }
    this.top -= 1;
    this.mode = this.top == 0
        ? 'd'
        : this.stack[this.top - 1] == null
            ? 'a'
            : 'k';
  }

  /**
   * Push an beginArray or beginObject scope.
   *
   * @param jo The scope to open.
   * @throws JsonException If nesting is too deep.
   */
  private void push(JsonObject jo) throws JsonException {
    if (this.top >= maxdepth) {
      throw new JsonException("Nesting too deep.");
    }
    this.stack[this.top] = jo;
    this.mode = jo == null ? 'a' : 'k';
    this.top += 1;
  }

  /**
   * Make a JSON text of an Object value. If the beginObject has an value.toJsonString() method,
 then that method will be used to produce the JSON text. The method is required to produce a
 strictly conforming text. If the beginObject does not contain a toJsonString method (which is
 the most common case), then a text will be produced by other means. If the value is an
 beginArray or Collection, then a JsonArray will be made from it and its toJsonString method
 will be called. If the value is a MAP, then a JsonObject will be made from it and its
 toJsonString method will be called. Otherwise, the value's toString method will be called, and
 the result will be quoted.

 <p>
   * Warning: This method assumes that the data structure is acyclical.
   *
   * @param value The value to be serialized.
   * @return a printable, displayable, transmittable representation of the beginObject, beginning
   * with
   * <code>{</code>&nbsp;<small>(left brace)</small> and ending with
   * <code>}</code>&nbsp;<small>(right brace)</small>.
   * @throws JsonException If the value is or contains an invalid number.
   */
  public static String valueToString(Object value) throws JsonException {
    if (value == null || value.equals(null)) {
      return "null";
    }
    if (value instanceof JsonString) {
      Object object;
      try {
        object = ((JsonString) value).toJsonString();
      } catch (Exception e) {
        throw new JsonException(e);
      }
      if (object instanceof String) {
        return (String) object;
      }
      throw new JsonException("Bad value from toJSONString: " + object);
    }
    if (value instanceof Number) {
      // not all Numbers may match actual JSON Numbers. i.e. Fractions or Complex
      final String numberAsString = JsonObject.numberToString((Number) value);
      try {
        // Use the BigDecimal constructor for it's parser to validate the format.
        @SuppressWarnings("unused")
        BigDecimal unused = new BigDecimal(numberAsString);
        // Close enough to a JSON number that we will return it unquoted
        return numberAsString;
      } catch (NumberFormatException ex) {
        // The Number value is not a valid JSON number.
        // Instead we will quote it as a string
        return JsonObject.quote(numberAsString);
      }
    }
    if (value instanceof Boolean || value instanceof JsonObject
        || value instanceof JsonArray) {
      return value.toString();
    }
    if (value instanceof Map) {
      Map<?, ?> map = (Map<?, ?>) value;
      return new JsonObject(map).toString();
    }
    if (value instanceof Collection) {
      Collection<?> coll = (Collection<?>) value;
      return new JsonArray(coll).toString();
    }
    if (value.getClass().isArray()) {
      return new JsonArray(value).toString();
    }
    if (value instanceof Enum<?>) {
      return JsonObject.quote(((Enum<?>) value).name());
    }
    return JsonObject.quote(value.toString());
  }

  /**
   * Append either the value <code>true</code> or the value <code>false</code>.
   *
   * @param b A boolean.
   * @return this
   * @throws JsonException
   */
  public JsonWriter value(boolean b) throws JsonException {
    return this.append(b ? "true" : "false");
  }

  /**
   * Append a double value.
   *
   * @param d A double.
   * @return this
   * @throws JsonException If the number is not finite.
   */
  public JsonWriter value(double d) throws JsonException {
    return this.value(new Double(d));
  }

  /**
   * Append a long value.
   *
   * @param l A long.
   * @return this
   * @throws JsonException
   */
  public JsonWriter value(long l) throws JsonException {
    return this.append(Long.toString(l));
  }

  /**
   * Append an beginObject value.
   *
   * @param object The beginObject to append. It can be null, or a Boolean, Number, String,
   * JsonObject, or JsonArray, or an beginObject that implements JsonString.
   * @return this
   * @throws JsonException If the value is out of sequence.
   */
  public JsonWriter value(Object object) throws JsonException {
    return this.append(valueToString(object));
  }
}
