/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * This file is part of terrayazilim-json project.
 *
 * This file incorporates work covered by
 * the following copyright and permission notices:
 *
 * Copyright (C) 2018 Terra Yazılım Bilişim Hiz. Elek. Dan. Oto. ve Loj. Tic. Ltd. Şti. | info [at] terrayazilim [dot] com [dot] tr
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package tr.com.terrayazilim.json.type;

import java.io.Serializable;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.GenericArrayType;
import java.lang.reflect.GenericDeclaration;
import java.lang.reflect.Modifier;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;
import java.lang.reflect.WildcardType;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Objects;
import tr.com.terrayazilim.json.JsonException;

/**
 * Derived and modified version of Gson.TypeToken.
 *
 * @author M.Çağrı Tepebaşılı - cagritepebasili [at] protonmail [dot] com
 * @version 1.0.3
 * @since 1.0.3
 * @param <T>
 */
public class Token<T> implements JavaToken<T> {

  private final Type type;
  private final Class<? super T> rawType;
  private final int hashCode;

  public Token() {
    this.type = getSuperParameterType(getClass());
    this.rawType = (Class<? super T>) rawTypeOf(this.type);
    this.hashCode = type.hashCode();
  }

  /**
   * @param type
   */
  public Token(Type type) {
    this.type = Token.localize(type);
    this.rawType = (Class<? super T>) rawTypeOf(this.type);
    this.hashCode = this.type.hashCode();
  }

  /**
   * @param <T>
   * @param type
   * @return
   */
  public static <T> Token<T> of(Type type) {
    return new Token<T>(type);
  }

  /**
   * @param <T>
   * @param type
   * @return
   */
  public static <T> Token<T> of(Class<T> type) {
    return new Token<T>(type);
  }
  
  @Override
  public Type getType() {
    return type;
  }

  @Override
  public Class<? super T> getRawType() {
    return rawType;
  }

  @Override
  public boolean isInterface() {
    return getRawType().isInterface();
  }

  @Override
  public boolean isAbstract() {
    return Modifier.isAbstract(getClass().getModifiers());
  }

  @Override
  public boolean isEnum() {
    return getRawType().isEnum();
  }

  @Override
  public boolean isArray() {
    Type tokenType = getType();

    return tokenType instanceof GenericArrayType
        || tokenType instanceof Class
        && ((Class<?>) tokenType).isArray();
  }
  
  /**
   * @return {@code true} if {@link #type} is instance of {@link java.lang.reflect.ParameterizedType}.
   * Otherwise {@code false}.
   */
  public boolean isParameterizedType() {
    return getType() instanceof ParameterizedType;
  }
  
  /**
   * @return {@code true} if {@link #type} is instance of {@link java.lang.reflect.TypeVariable}.
   * Otherwise {@code false}.
   */
  public boolean isTypeVariable() {
    return getType() instanceof TypeVariable;
  }
  
  /**
   * @return {@code true} if {@link #type} is instance of {@link java.lang.reflect.WildcardType}.
   * Otherwise {@code false}.
   */
  public boolean isWildcardType() {
    return getType() instanceof WildcardType;
  }
  
  public boolean isCollection() {
    Class<? super T> raw = getRawType();
    
    return Collection.class.isAssignableFrom(raw);
  }

  public boolean isMap() {
    Class<? super T> raw = getRawType();
    
    return Map.class.isAssignableFrom(raw);
  }
  
  /**
   * @return 
   */
  public Collection<FieldToken> getDeclaredFields() {
    return getDeclaredFields(getRawType());
  }
  
  /**
   * @param type
   * @return 
   */
  public static Collection<FieldToken> getDeclaredFields(Class<?> type) {
    Collection<FieldToken> tokens = new ArrayList<>();

    Field[] fields = type.getDeclaredFields();
    if (tr.com.terrayazilim.core.Objects.isNull(fields) || Array.getLength(fields) == 0) {
      return tokens;
    }

    for (Field field : fields) {
      if (!field.getName().equals("this$0")) {
        FieldToken token = new FieldToken(field);
        tokens.add(token);
      }
    }

    return tokens;
  }
  
  public static Type resolveBounds(WildcardType wt) {
    if (Array.getLength(wt.getLowerBounds()) > 0) {
      return wt.getLowerBounds()[0];
    } else if (Array.getLength(wt.getUpperBounds()) > 0) {
      return wt.getUpperBounds()[0];
    } else {
      return Object.class;
    }
  }
  
  public static Type resolveWildCardType(Type sign) {
    Type resolvedType;

    if (sign instanceof WildcardType) {
      WildcardType wt = (WildcardType) sign;
      resolvedType = Token.resolveBounds(wt);

      if ((resolvedType != null) && (resolvedType instanceof Class<?>)) {
        return resolvedType;
      }
    }

    return sign;
  }
  
  public static Type getType(Type context, Class<?> contextRawType) {
    Type collectionType = Token.getSupertype(context, contextRawType, Collection.class);

    if (collectionType instanceof WildcardType) {
      collectionType = ((WildcardType) collectionType).getUpperBounds()[0];
    }
    if (collectionType instanceof ParameterizedType) {
      return ((ParameterizedType) collectionType).getActualTypeArguments()[0];
    }

    return Object.class;
  }

  /**
   * @param toResolve
   * @return 
   */
  public Type getGenericSuperType(Class<?> toResolve) {
    return getGenericSuperType(getType(), getRawType(), toResolve);
  }
  
  public static Type getGenericSuperType(Type contextType, Class<?> rawType, Class<?> resolved) {
    if (resolved == rawType) {
      return contextType;
    }

    if (resolved.isInterface()) {
      Class<?>[] interfaces = rawType.getInterfaces();
      for (int i = 0, length = interfaces.length; i < length; i++) {
        if (interfaces[i] == resolved) {
          return rawType.getGenericInterfaces()[i];
        } else if (resolved.isAssignableFrom(interfaces[i])) {
          return getGenericSuperType(rawType.getGenericInterfaces()[i], interfaces[i], resolved);
        }
      }
    }

    if (!rawType.isInterface()) {
      while (rawType != Object.class) {
        Class<?> rawSupertype = rawType.getSuperclass();
        if (rawSupertype == resolved) {
          return rawType.getGenericSuperclass();
        } else if (resolved.isAssignableFrom(rawSupertype)) {
          return getGenericSuperType(rawType.getGenericSuperclass(), rawSupertype, resolved);
        }
        rawType = rawSupertype;
      }
    }

    return resolved;
  }
  
  /**
   * @param toResolve
   * @return 
   */
  public Type resolve(Type toResolve) {
    return resolve(getType(), getRawType(), toResolve);
  }

  public static Type resolve(Type contextType, Class<?> contextRawType, Type toResolve) {
    while (true) {
      if (toResolve instanceof TypeVariable) {
        TypeVariable<?> typeVariable = (TypeVariable<?>) toResolve;
        toResolve = Token.resolveTypeVariable(contextType, contextRawType, typeVariable);
        if (toResolve == typeVariable) {
          return toResolve;
        }

      } else if (toResolve instanceof Class && ((Class<?>) toResolve).isArray()) {
        Class<?> original = (Class<?>) toResolve;
        Type componentType = original.getComponentType();
        Type newComponentType = Token.resolve(contextType, contextRawType, componentType);
        return componentType == newComponentType
            ? original
            : arrayOf(newComponentType);

      } else if (toResolve instanceof GenericArrayType) {
        GenericArrayType original = (GenericArrayType) toResolve;
        Type componentType = original.getGenericComponentType();
        Type newComponentType = Token.resolve(contextType, contextRawType, componentType);
        return componentType == newComponentType
            ? original
            : arrayOf(newComponentType);

      } else if (toResolve instanceof ParameterizedType) {
        ParameterizedType original = (ParameterizedType) toResolve;
        Type ownerType = original.getOwnerType();
        Type newOwnerType = Token.resolve(contextType, contextRawType, ownerType);
        boolean changed = newOwnerType != ownerType;

        Type[] args = original.getActualTypeArguments();
        for (int t = 0, length = args.length; t < length; t++) {
          Type resolvedTypeArgument = Token.resolve(contextType, contextRawType, args[t]);
          if (resolvedTypeArgument != args[t]) {
            if (!changed) {
              args = args.clone();
              changed = true;
            }
            args[t] = resolvedTypeArgument;
          }
        }

        return changed
            ? newParameterizedTypeWithOwner(newOwnerType, original.getRawType(), args)
            : original;

      } else if (toResolve instanceof WildcardType) {
        WildcardType original = (WildcardType) toResolve;
        Type[] originalLowerBound = original.getLowerBounds();
        Type[] originalUpperBound = original.getUpperBounds();

        if (originalLowerBound.length == 1) {
          Type lowerBound = Token.resolve(contextType, contextRawType, originalLowerBound[0]);
          if (lowerBound != originalLowerBound[0]) {
            return supertypeOf(lowerBound);
          }
        } else if (originalUpperBound.length == 1) {
          Type upperBound = Token.resolve(contextType, contextRawType, originalUpperBound[0]);
          if (upperBound != originalUpperBound[0]) {
            return subtypeOf(upperBound);
          }
        }
        return original;

      } else {
        return toResolve;
      }
    }
  }

  /**
   * @param ownerType
   * @param rawType
   * @param typeArguments
   * @return
   */
  public static ParameterizedType newParameterizedTypeWithOwner(
      Type ownerType, Type rawType, Type... typeArguments) {
    return new ParameterizedTypeImpl(ownerType, rawType, typeArguments);
  }

  /**
   * @param componentType
   * @return a {@link java.io.Serializable serializable} generic array type.
   */
  public static GenericArrayType arrayOf(Type componentType) {
    return new GenericArrayTypeImpl(componentType);
  }

  /**
   * @param bound
   * @return
   */
  public static WildcardType subtypeOf(Type bound) {
    return new WildcardTypeImpl(new Type[]{bound}, new WildcardType[0]);
  }

  /**
   * @param bound
   * @return
   */
  public static WildcardType supertypeOf(Type bound) {
    return new WildcardTypeImpl(new Type[]{Object.class}, new Type[]{bound});
  }
  
  /**
   * @param typeVariable
   * @return 
   */
  public Type resolveTypeVariable(TypeVariable typeVariable) {
    return resolveTypeVariable(getType(), getRawType(), typeVariable);
  }

  /**
   * @param context
   * @param contextRawType
   * @param unknown
   * @return
   */
  public static Type resolveTypeVariable(Type context, Class<?> contextRawType, TypeVariable<?> unknown) {
    Class<?> declaredByRaw = declaringClassOf(unknown);

    if (declaredByRaw == null) {
      return unknown;
    }

    Type declaredBy = getGenericSuperType(context, contextRawType, declaredByRaw);
    if (declaredBy instanceof ParameterizedType) {
      int index = indexOf(declaredByRaw.getTypeParameters(), unknown);
      return ((ParameterizedType) declaredBy).getActualTypeArguments()[index];
    }

    return unknown;
  }

  private static int indexOf(Object[] array, Object toFind) {
    for (int i = 0, length = array.length; i < length; i++) {
      if (toFind.equals(array[i])) {
        return i;
      }
    }
    throw new NoSuchElementException();
  }

  public static Class<?> declaringClassOf(TypeVariable<?> typeVariable) {
    GenericDeclaration genericDeclaration = typeVariable.getGenericDeclaration();
    return genericDeclaration instanceof Class
        ? (Class<?>) genericDeclaration
        : null;
  }

  /**
   * @param type
   * @return
   */
  public static Type localize(Type type) {
    if (type instanceof Class) {
      Class<?> c = (Class<?>) type;
      return c.isArray() ? new GenericArrayTypeImpl(Token.localize(c.getComponentType())) : c;

    } else if (type instanceof ParameterizedType) {
      ParameterizedType p = (ParameterizedType) type;
      return new ParameterizedTypeImpl(p.getOwnerType(),
          p.getRawType(), p.getActualTypeArguments());

    } else if (type instanceof GenericArrayType) {
      GenericArrayType g = (GenericArrayType) type;
      return new GenericArrayTypeImpl(g.getGenericComponentType());

    } else if (type instanceof WildcardType) {
      WildcardType w = (WildcardType) type;
      return new WildcardTypeImpl(w.getUpperBounds(), w.getLowerBounds());

    } else {
      return type;
    }
  }

  /**
   * @param type
   * @return
   */
  public static Class<?> rawTypeOf(Type type) {
    if (type instanceof Class<?>) {
      return (Class<?>) type;
    } else if (type instanceof ParameterizedType) {
      ParameterizedType parameterizedType = (ParameterizedType) type;
      Type rawType = parameterizedType.getRawType();
      return (Class<?>) rawType;
    } else if (type instanceof GenericArrayType) {
      Type componentType = ((GenericArrayType) type).getGenericComponentType();
      return Array.newInstance(rawTypeOf(componentType), 0).getClass();
    } else if (type instanceof TypeVariable) {
      return Object.class;
    } else if (type instanceof WildcardType) {
      return rawTypeOf(((WildcardType) type).getUpperBounds()[0]);
    } else {
      throw new JsonException("Unresolved Type. {Hint: ErrorCode:119}");
    }
  }

  /**
   * @param context
   * @param contextRawType
   * @param supertype
   * @return
   */
  public static Type getSupertype(Type context, Class<?> contextRawType, Class<?> supertype) {
    return Token.resolve(context, contextRawType,
        getGenericSuperType(context, contextRawType, supertype));
  }

  /**
   * @param subclass
   * @return
   */
  public Type getSuperParameterType(Class<?> subclass) {
    Type sub = subclass.getGenericSuperclass();
    ParameterizedType pt = (ParameterizedType) sub;

    return pt.getActualTypeArguments()[0];
  }

  @Override
  public boolean equals(Object o) {
    return o instanceof Token<?> && equals(type, ((Token<?>) o).type);
  }

  @Override
  public int hashCode() {
    return this.hashCode;
  }

  /**
   * @param a
   * @param b
   * @return
   */
  private static boolean equal(Object a, Object b) {
    return a == b || (a != null && a.equals(b));
  }

  /**
   * @param a
   * @param b
   * @return
   */
  public static boolean equals(Type a, Type b) {
    if (a == b) {
      return true;
    } else if (a instanceof Class) {
      return a.equals(b);
    } else if (a instanceof ParameterizedType) {
      if (!(b instanceof ParameterizedType)) {
        return false;
      }
      ParameterizedType pa = (ParameterizedType) a;
      ParameterizedType pb = (ParameterizedType) b;
      return equal(pa.getOwnerType(), pb.getOwnerType())
          && pa.getRawType().equals(pb.getRawType())
          && Arrays.equals(pa.getActualTypeArguments(), pb.getActualTypeArguments());

    } else if (a instanceof GenericArrayType) {
      if (!(b instanceof GenericArrayType)) {
        return false;
      }
      GenericArrayType ga = (GenericArrayType) a;
      GenericArrayType gb = (GenericArrayType) b;
      return equals(ga.getGenericComponentType(), gb.getGenericComponentType());

    } else if (a instanceof WildcardType) {
      if (!(b instanceof WildcardType)) {
        return false;
      }
      WildcardType wa = (WildcardType) a;
      WildcardType wb = (WildcardType) b;
      return Arrays.equals(wa.getUpperBounds(), wb.getUpperBounds())
          && Arrays.equals(wa.getLowerBounds(), wb.getLowerBounds());

    } else if (a instanceof TypeVariable) {
      if (!(b instanceof TypeVariable)) {
        return false;
      }
      TypeVariable<?> va = (TypeVariable<?>) a;
      TypeVariable<?> vb = (TypeVariable<?>) b;
      return va.getGenericDeclaration() == vb.getGenericDeclaration()
          && va.getName().equals(vb.getName());

    } else {
      return false;
    }
  }

  /**
   * @version 1.0.3
   * @since 1.0.3
   */
  static final class ParameterizedTypeImpl implements ParameterizedType, Serializable {

    private final Type ownerType;
    private final Type rawType;
    private final Type[] typeArguments;

    /**
     * @param ownerType
     * @param rawType
     * @param typeArguments
     */
    public ParameterizedTypeImpl(Type ownerType, Type rawType, Type[] typeArguments) {
      this.ownerType = ownerType == null ? null : Token.localize(ownerType);
      this.rawType = Token.localize(rawType);
      this.typeArguments = typeArguments.clone();

      for (int t = 0, length = this.typeArguments.length; t < length; t++) {
        this.typeArguments[t] = Token.localize(this.typeArguments[t]);
      }
    }

    @Override
    public Type[] getActualTypeArguments() {
      return this.typeArguments;
    }

    @Override
    public Type getRawType() {
      return this.rawType;
    }

    @Override
    public Type getOwnerType() {
      return this.ownerType;
    }

    @Override
    public boolean equals(Object other) {
      return other instanceof ParameterizedType && Token.equals(this, (ParameterizedType) other);
    }

    @Override
    public int hashCode() {
      int hash = 3;
      hash = 53 * hash + Objects.hashCode(this.ownerType);
      hash = 53 * hash + Objects.hashCode(this.rawType);
      hash = 53 * hash + Arrays.deepHashCode(this.typeArguments);
      return hash;
    }
  }

  /**
   * @version 1.0.3
   * @since 1.0.3
   */
  static final class GenericArrayTypeImpl implements GenericArrayType, Serializable {

    private final Type componentType;

    /**
     * @param componentType
     */
    public GenericArrayTypeImpl(Type componentType) {
      this.componentType = localize(componentType);
    }

    @Override
    public Type getGenericComponentType() {
      return this.componentType;
    }

    @Override
    public boolean equals(Object o) {
      return o instanceof GenericArrayType && Token.equals(this, (GenericArrayType) o);
    }

    @Override
    public int hashCode() {
      return componentType.hashCode();
    }
  }

  /**
   * @version 1.0.3
   * @since 1.0.0
   */
  static final class WildcardTypeImpl implements WildcardType, Serializable {

    private final Type upperBound;
    private final Type lowerBound;

    /**
     * @param upperBound
     * @param lowerBound
     */
    public WildcardTypeImpl(Type[] upperBound, Type[] lowerBound) {
      if (lowerBound.length == 1) {
        this.lowerBound = localize(lowerBound[0]);
        this.upperBound = Object.class;
      } else {
        this.lowerBound = null;
        this.upperBound = localize(upperBound[0]);
      }
    }

    @Override
    public boolean equals(Object other) {
      return other instanceof WildcardType && Token.equals(this, (WildcardType) other);
    }

    @Override
    public int hashCode() {
      int hash = 5;
      hash = 37 * hash + Objects.hashCode(this.upperBound);
      hash = 37 * hash + Objects.hashCode(this.lowerBound);
      return hash;
    }

    @Override
    public Type[] getUpperBounds() {
      return new Type[]{upperBound};
    }

    @Override
    public Type[] getLowerBounds() {
      return (lowerBound != null)
          ? new Type[]{lowerBound}
          : new Type[]{};
    }
  }
}
